# type: ignore
# BitBIRCH is an open-source clustering module based on iSIM
#
# Please, cite the BitBIRCH paper: https://doi.org/10.1039/D5DD00030K
#
# BitBIRCH is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# BitBIRCH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# BitBIRCH License: GPL-3.0 https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Memory-efficient BitBIRCH authors: Ramon Alain Miranda Quintana <ramirandaq@gmail.com>, <quintana@chem.ufl.edu>
#                                    Krizstina Zsigmond <kzsigmond@ufl.edu>
#
### Part of the tree-management code was derived from https://scikit-learn.org/stable/modules/generated/sklearn.cluster.Birch.html
### Authors: Manoj Kumar <manojkumarsivaraj334@gmail.com>
###          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
###          Joel Nothman <joel.nothman@gmail.com>
### License: BSD 3 clause
# Parts of the BitBIRCH algorithm were previously released under the LGPL-3.0 license by:
# Ramon Alain Miranda Quintana <ramirandaq@gmail.com>, <quintana@chem.ufl.edu>
# Vicky (Vic) Jung <jungvicky@ufl.edu>
# Kenneth Lopez Perez <klopezperez@chem.ufl.edu>
# Kate Huddleston <kdavis2@chem.ufl.edu>
from pathlib import Path

import numpy as np
from scipy import sparse


def safe_sum(nmax, np1, np2):
    if nmax >= 4294967294:
        return np1.astype("uint64") + np2.astype("uint64")
    elif nmax >= 65534:
        return np1.astype("uint32") + np2.astype("uint32")
    elif nmax >= 254:
        return np1.astype("uint16") + np2.astype("uint16")
    else:
        return np1 + np2


def _copy_or_unpack(x, n_features, input_is_packed: bool = True):
    return unpack_fingerprints(x, n_features) if input_is_packed else x.copy()


def unpack_fingerprints(a, n_features: int):
    """Unpacks uint8 arrays into boolean arrays"""
    # n_features is required to discard padded zeros if it is not a multiple of 8
    return np.unpackbits(a, axis=-1, count=n_features)


# Utility function to validate the n_features argument for packed inputs
def _validate_n_features(X, input_is_packed: bool, n_features: int | None) -> int:
    if input_is_packed:
        if n_features is None:
            raise ValueError("n_features is required for packed inputs")
        return n_features

    x_n_features = X.shape[1]
    if n_features is not None:
        if n_features != x_n_features:
            raise ValueError(
                "n_features is redundant for non-packed inputs"
                " if passed, it must be equal to X.shape[1]."
                f" For passed X, X.shape[1] = {X.shape[1]}."
                " If this value is not what you expected,"
                " make sure the passed X is actually unpacked."
            )
    return x_n_features


def set_merge(merge_criterion, tolerance=0.05):
    """
    Sets merge_accept function for merge_subcluster, based on user specified merge_criteria.

    Radius: merge subcluster based on comparison to centroid of the cluster
    Diameter: merge subcluster based on instant Tanimoto similarity of cluster
    Tolerance: applies tolerance threshold to diameter merge criteria, which will merge subcluster with stricter threshold for newly added molecules

    Parameters:
    -----------
    merge_criterion: str();
                        radius, diameter or tolerance
    tolerance: float;
                        sets penalty value for similarity threshold when callng tolerance merge criteria

    Returns:
    --------
    merge_accept(): function
                        if cluster is accepted to merge, merge the cluster based on the criteria specified
    """
    if merge_criterion == "radius":

        def merge_accept(
            threshold, new_ls, new_centroid, new_n, old_ls, nom_ls, old_n, nom_n
        ):
            jt_sim = jt_isim(new_ls + new_centroid, new_n + 1) * (new_n + 1) - jt_isim(
                new_ls, new_n
            ) * (new_n - 1)
            return jt_sim >= threshold * 2

    elif merge_criterion == "diameter":

        def merge_accept(
            threshold, new_ls, new_centroid, new_n, old_ls, nom_ls, old_n, nom_n
        ):
            jt_radius = jt_isim(new_ls, new_n)
            return jt_radius >= threshold

    elif merge_criterion == "tolerance_tough":

        def merge_accept(
            threshold, new_ls, new_centroid, new_n, old_ls, nom_ls, old_n, nom_n
        ):
            jt_radius = jt_isim(new_ls, new_n)
            if jt_radius < threshold:
                return False
            else:
                if old_n == 1 and nom_n == 1:
                    return True
                elif nom_n == 1:
                    return (
                        jt_isim(old_ls + nom_ls, old_n + 1) * (old_n + 1)
                        - jt_isim(old_ls, old_n) * (old_n - 1)
                    ) / 2 >= jt_isim(old_ls, old_n) - tolerance and (
                        jt_radius >= threshold
                    )
                else:
                    return (
                        jt_isim(old_ls + nom_ls, old_n + nom_n)
                        * (old_n + nom_n)
                        * (old_n + nom_n - 1)
                        - jt_isim(old_ls, old_n) * old_n * (old_n - 1)
                        - jt_isim(nom_ls, nom_n) * nom_n * (nom_n - 1)
                    ) / (2 * old_n * nom_n) >= jt_isim(old_ls, old_n) - tolerance and (
                        jt_radius >= threshold
                    )

    elif merge_criterion in ["tolerance", "tolerance-legacy"]:

        def merge_accept(
            threshold, new_ls, new_centroid, new_n, old_ls, nom_ls, old_n, nom_n
        ):
            jt_radius = jt_isim(new_ls, new_n)
            if jt_radius < threshold:
                return False
            else:
                if old_n == 1 and nom_n == 1:
                    return True
                elif nom_n == 1:
                    return (
                        jt_isim(old_ls + nom_ls, old_n + 1) * (old_n + 1)
                        - jt_isim(old_ls, old_n) * (old_n - 1)
                    ) / 2 >= jt_isim(old_ls, old_n) - tolerance and (
                        jt_radius >= threshold
                    )
                else:
                    return True

    globals()["merge_accept"] = merge_accept


def jt_isim(c_total, n_objects):
    """iSIM Tanimoto calculation

    https://pubs.rsc.org/en/content/articlelanding/2024/dd/d4dd00041b

    Parameters
    ----------
    c_total : np.ndarray
              Sum of the elements column-wise

    n_objects : int
                Number of elements

    Returns
    ----------
    isim : float
           iSIM Jaccard-Tanimoto value
    """
    x = c_total.astype("uint64")
    sum_kq = np.sum(x)
    sum_kqsq = np.dot(x, x)
    a = (sum_kqsq - sum_kq) / 2
    return a / (a + n_objects * sum_kq - sum_kqsq)


def max_separation(Y):
    """Finds two objects in X that are very separated
    This is an approximation (not guaranteed to find
    the two absolutely most separated objects), but it is
    a very robust O(N) implementation. Quality of clustering
    does not diminish in the end.

    Algorithm:
    a) Find centroid of X
    b) mol1 is the molecule most distant from the centroid
    c) mol2 is the molecule most distant from mol1

    Returns
    -------
    (mol1, mol2) : (int, int)
                   indices of mol1 and mol2
    1 - sims_mol1 : np.ndarray
                   Distances to mol1
    1 - sims_mol2: np.ndarray
                   Distances to mol2
    These are needed for node1_dist and node2_dist in _split_node
    """
    # Get the centroid of the set
    X = Y.astype("uint64")
    n_samples = len(X)
    linear_sum = np.sum(X, axis=0)
    centroid = calc_centroid(linear_sum, n_samples)

    # Get the similarity of each molecule to the centroid
    pop_counts = np.sum(X, axis=1)
    a_centroid = np.dot(X, centroid)
    sims_med = a_centroid / (pop_counts + np.sum(centroid) - a_centroid)

    # Get the least similar molecule to the centroid
    mol1 = np.argmin(sims_med)

    # Get the similarity of each molecule to mol1
    a_mol1 = np.dot(X, X[mol1])
    sims_mol1 = a_mol1 / (pop_counts + pop_counts[mol1] - a_mol1)

    # Get the least similar molecule to mol1
    mol2 = np.argmin(sims_mol1)

    # Get the similarity of each molecule to mol2
    a_mol2 = np.dot(X, X[mol2])
    sims_mol2 = a_mol2 / (pop_counts + pop_counts[mol2] - a_mol2)

    return (mol1, mol2), sims_mol1, sims_mol2


def calc_centroid(linear_sum, n_samples):
    """Calculates centroid

    Parameters
    ----------

    linear_sum : np.ndarray
                 Sum of the elements column-wise
    n_samples : int
                Number of samples

    Returns
    -------
    centroid : np.ndarray
               Centroid fingerprints of the given set
    """
    cent = np.where(linear_sum >= n_samples * 0.5, 1, 0)
    return cent.astype("bool")


def _iterate_sparse_X(X):
    """This little hack returns a densified row when iterating over a sparse
    matrix, instead of constructing a sparse matrix for every row that is
    expensive.
    """
    n_samples, n_features = X.shape
    X_indices = X.indices
    X_data = X.data
    X_indptr = X.indptr

    for i in range(n_samples):
        row = np.zeros(n_features)
        startptr, endptr = X_indptr[i], X_indptr[i + 1]
        nonzero_indices = X_indices[startptr:endptr]
        row[nonzero_indices] = X_data[startptr:endptr]
        yield row


def _split_node(node, threshold, branching_factor):
    """The node has to be split if there is no place for a new subcluster
    in the node.
    1. Two empty nodes and two empty subclusters are initialized.
    2. The pair of distant subclusters are found.
    3. The properties of the empty subclusters and nodes are updated
       according to the nearest distance between the subclusters to the
       pair of distant subclusters.
    4. The two nodes are set as children to the two subclusters.
    """
    new_subcluster1 = _BFSubcluster()
    new_subcluster2 = _BFSubcluster()
    new_node1 = _BFNode(
        threshold=threshold,
        branching_factor=branching_factor,
        is_leaf=node.is_leaf,
        n_features=node.n_features,
        dtype=node.init_centroids_.dtype,
    )
    new_node2 = _BFNode(
        threshold=threshold,
        branching_factor=branching_factor,
        is_leaf=node.is_leaf,
        n_features=node.n_features,
        dtype=node.init_centroids_.dtype,
    )
    new_subcluster1.child_ = new_node1
    new_subcluster2.child_ = new_node2

    if node.is_leaf:
        if node.prev_leaf_ is not None:
            node.prev_leaf_.next_leaf_ = new_node1
        new_node1.prev_leaf_ = node.prev_leaf_
        new_node1.next_leaf_ = new_node2
        new_node2.prev_leaf_ = new_node1
        new_node2.next_leaf_ = node.next_leaf_
        if node.next_leaf_ is not None:
            node.next_leaf_.prev_leaf_ = new_node2

    # O(N) implementation of max separation
    farthest_idx, node1_dist, node2_dist = max_separation(node.centroids_)
    # Notice that max_separation is returning similarities and not distances
    node1_closer = node1_dist > node2_dist
    # Make sure node1 is closest to itself even if all distances are equal.
    # This can only happen when all node.centroids_ are duplicates leading to all
    # distances between centroids being zero.
    node1_closer[farthest_idx[0]] = True

    for idx, subcluster in enumerate(node.subclusters_):
        if node1_closer[idx]:
            new_node1.append_subcluster(subcluster)
            new_subcluster1.update(subcluster)
            # if not singly:
            #    subcluster.parent_ = new_subcluster1
        else:
            new_node2.append_subcluster(subcluster)
            new_subcluster2.update(subcluster)
            # if not singly:
            #    subcluster.parent_ = new_subcluster2
    return new_subcluster1, new_subcluster2


class _BFNode:
    """Each node in a BFTree is called a BFNode.

    The BFNode can have a maximum of branching_factor
    number of BFSubclusters.

    Parameters
    ----------
    threshold : float
        Threshold needed for a new subcluster to enter a BFSubcluster.

    branching_factor : int
        Maximum number of BF subclusters in each node.

    is_leaf : bool
        We need to know if the BFNode is a leaf or not, in order to
        retrieve the final subclusters.

    n_features : int
        The number of features.

    Attributes
    ----------
    subclusters_ : list
        List of subclusters for a particular BFNode.

    prev_leaf_ : _BFNode
        Useful only if is_leaf is True.

    next_leaf_ : _BFNode
        next_leaf. Useful only if is_leaf is True.
        the final subclusters.

    init_centroids_ : ndarray of shape (branching_factor + 1, n_features)
        Manipulate ``init_centroids_`` throughout rather than centroids_ since
        the centroids are just a view of the ``init_centroids_`` .

    centroids_ : ndarray of shape (branching_factor + 1, n_features)
        View of ``init_centroids_``.

    """

    def __init__(self, *, threshold, branching_factor, is_leaf, n_features, dtype):
        self.threshold = threshold
        self.branching_factor = branching_factor
        self.is_leaf = is_leaf
        self.n_features = n_features

        # The list of subclusters, centroids and squared norms
        # to manipulate throughout.
        self.subclusters_ = []
        self.init_centroids_ = np.zeros((branching_factor + 1, n_features), dtype=dtype)
        self.prev_leaf_ = None
        self.next_leaf_ = None

    def append_subcluster(self, subcluster):
        n_samples = len(self.subclusters_)
        self.subclusters_.append(subcluster)
        self.init_centroids_[n_samples] = subcluster.centroid_

        # Keep centroids as views. In this way
        # if we change init_centroids, it is sufficient
        self.centroids_ = self.init_centroids_[: n_samples + 1, :]

    def update_split_subclusters(self, subcluster, new_subcluster1, new_subcluster2):
        """Remove a subcluster from a node and update it with the
        split subclusters.
        """

        ind = self.subclusters_.index(subcluster)
        self.subclusters_[ind] = new_subcluster1
        self.init_centroids_[ind] = new_subcluster1.centroid_
        self.centroids_[ind] = new_subcluster1.centroid_
        self.append_subcluster(new_subcluster2)

    def insert_bf_subcluster(self, subcluster, set_bits):
        """Insert a new subcluster into the node."""
        if not self.subclusters_:
            self.append_subcluster(subcluster)
            return False

        threshold = self.threshold
        branching_factor = self.branching_factor
        # We need to find the closest subcluster among all the
        # subclusters so that we can insert our new subcluster.
        sub_centroids = self.centroids_.astype("uint16")
        in_centroid = subcluster.centroid_.astype("uint16")
        a = np.dot(sub_centroids, in_centroid)
        sim_matrix = a / (np.sum(sub_centroids, axis=1) + set_bits - a)
        closest_index = np.argmax(sim_matrix)
        closest_subcluster = self.subclusters_[closest_index]

        # If the subcluster has a child, we need a recursive strategy.
        if closest_subcluster.child_ is not None:

            split_child = closest_subcluster.child_.insert_bf_subcluster(
                subcluster, set_bits
            )

            if not split_child:
                # If it is determined that the child need not be split, we
                # can just update the closest_subcluster
                closest_subcluster.update(subcluster)
                self.init_centroids_[closest_index] = self.subclusters_[
                    closest_index
                ].centroid_
                self.centroids_[closest_index] = self.subclusters_[
                    closest_index
                ].centroid_
                return False

            # things not too good. we need to redistribute the subclusters in
            # our child node, and add a new subcluster in the parent
            # subcluster to accommodate the new child.
            else:
                new_subcluster1, new_subcluster2 = _split_node(
                    closest_subcluster.child_, threshold, branching_factor
                )
                self.update_split_subclusters(
                    closest_subcluster, new_subcluster1, new_subcluster2
                )

                if len(self.subclusters_) > self.branching_factor:
                    return True
                return False

        # good to go!
        else:
            merged = closest_subcluster.merge_subcluster(subcluster, self.threshold)
            if merged:
                self.centroids_[closest_index] = closest_subcluster.centroid_
                self.init_centroids_[closest_index] = closest_subcluster.centroid_

                return False

            # not close to any other subclusters, and we still
            # have space, so add.
            elif len(self.subclusters_) < self.branching_factor:
                self.append_subcluster(subcluster)

                return False

            # We do not have enough space nor is it closer to an
            # other subcluster. We need to split.
            else:
                self.append_subcluster(subcluster)
                return True


class _BFSubcluster:
    """Each subcluster in a BFNode is called a BFSubcluster.

    A BFSubcluster can have a BFNode has its child.

    Parameters
    ----------
    linear_sum : ndarray of shape (n_features,), default=None
        Sample. This is kept optional to allow initialization of empty
        subclusters.

    Attributes
    ----------
    n_samples_ : int
        Number of samples that belong to each subcluster.

    linear_sum_ : ndarray
        Linear sum of all the samples in a subcluster. Prevents holding
        all sample data in memory.

    centroid_ : ndarray of shape (branching_factor + 1, n_features)
        Centroid of the subcluster. Prevent recomputing of centroids when
        ``BFNode.centroids_`` is called.

    mol_indices : list, default=[]
        List of indices of molecules included in the given cluster.

    child_ : _BFNode
        Child Node of the subcluster. Once a given _BFNode is set as the child
        of the _BFNode, it is set to ``self.child_``.
    """

    def __init__(self, *, linear_sum=None, mol_indices=[]):
        if linear_sum is None:
            self.n_samples_ = 0
            self.centroid_ = self.linear_sum_ = np.zeros((2048,), dtype="bool")
            self.mol_indices = []
        else:
            self.n_samples_ = 1
            self.centroid_ = self.linear_sum_ = linear_sum
            self.mol_indices = mol_indices

        self.child_ = None
        self.parent_ = None

    def update(self, subcluster):
        self.n_samples_ += np.uint64(subcluster.n_samples_)
        self.linear_sum_ = safe_sum(
            self.n_samples_, self.linear_sum_, subcluster.linear_sum_
        )
        self.mol_indices += subcluster.mol_indices
        self.centroid_ = calc_centroid(self.linear_sum_, self.n_samples_)

    def merge_subcluster(self, nominee_cluster, threshold):
        """Check if a cluster is worthy enough to be merged. If
        yes then merge.
        """
        new_n = np.uint64(self.n_samples_) + np.uint64(nominee_cluster.n_samples_)
        new_ls = safe_sum(new_n, self.linear_sum_, nominee_cluster.linear_sum_)
        new_centroid = calc_centroid(new_ls, new_n)

        if merge_accept(
            threshold,
            new_ls,
            new_centroid,
            new_n,
            self.linear_sum_,
            nominee_cluster.linear_sum_,
            self.n_samples_,
            nominee_cluster.n_samples_,
        ):
            (self.n_samples_, self.linear_sum_, self.centroid_, self.mol_indices) = (
                new_n,
                new_ls,
                new_centroid,
                self.mol_indices + nominee_cluster.mol_indices,
            )
            return True
        return False


class BitBirch:
    """Implements the BitBIRCH clustering algorithm.

    BitBIRCH paper:

    Memory- and time-efficient, online-learning algorithm.
    It constructs a tree data structure with the cluster centroids being read off the leaf.

    Parameters
    ----------
    threshold : float, default=0.5
        The similarity radius of the subcluster obtained by merging a new sample and the
        closest subcluster should be greater than the threshold. Otherwise a new
        subcluster is started. Setting this value to be very low promotes
        splitting and vice-versa.

    branching_factor : int, default=50
        Maximum number of BF subclusters in each node. If a new samples enters
        such that the number of subclusters exceed the branching_factor then
        that node is split into two nodes with the subclusters redistributed
        in each. The parent subcluster of that node is removed and two new
        subclusters are added as parents of the 2 split nodes.

    Attributes
    ----------
    root_ : _BFNode
        Root of the BFTree.

    dummy_leaf_ : _BFNode
        Start pointer to all the leaves.

    subcluster_centers_ : ndarray
        Centroids of all subclusters read directly from the leaves.

    Notes
    -----
    The tree data structure consists of nodes with each node consisting of
    a number of subclusters. The maximum number of subclusters in a node
    is determined by the branching factor. Each subcluster maintains a
    linear sum, mol_indices and the number of samples in that subcluster.
    In addition, each subcluster can also have a node as its child, if the
    subcluster is not a member of a leaf node.

    For a new point entering the root, it is merged with the subcluster closest
    to it and the linear sum, mol_indices and the number of samples of that
    subcluster are updated. This is done recursively till the properties of
    the leaf node are updated.
    """

    def __init__(
        self,
        *,
        threshold=0.5,
        branching_factor=50,
    ):
        self.threshold = threshold
        self.branching_factor = branching_factor
        self.index_tracker = 0
        self.first_call = True

    def fit(
        self,
        X,
        store_centroids: bool = False,
        input_is_packed: bool = True,
        n_features: int | None = None,
        max_fps: int | None = None,
    ):
        """
        Build a BF Tree for the input data.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Input data.

        Returns
        -------
        self
            Fitted estimator.
        """
        if isinstance(X, Path):
            X = np.load(X, mmap_mode="r")[:max_fps]
        else:
            X = X[:max_fps]
        threshold = self.threshold
        branching_factor = self.branching_factor
        n_features = _validate_n_features(X, input_is_packed, n_features)
        d_type = X.dtype

        # If partial_fit is called for the first time or fit is called, we
        # start a new tree.
        if self.first_call:
            # The first root is the leaf. Manipulate this object throughout.
            self.root_ = _BFNode(
                threshold=threshold,
                branching_factor=branching_factor,
                is_leaf=True,
                n_features=n_features,
                dtype=d_type,
            )

            # To enable getting back subclusters.
            self.dummy_leaf_ = _BFNode(
                threshold=threshold,
                branching_factor=branching_factor,
                is_leaf=True,
                n_features=n_features,
                dtype=d_type,
            )
            self.dummy_leaf_.next_leaf_ = self.root_
            self.root_.prev_leaf_ = self.dummy_leaf_

        # Cannot vectorize. Enough to convince to use cython.
        if not sparse.issparse(X):
            iter_func = iter
        else:
            iter_func = _iterate_sparse_X

        for sample in iter_func(X):
            unpack = _copy_or_unpack(sample, n_features, input_is_packed)
            set_bits = np.sum(unpack.astype("uint64"))
            subcluster = _BFSubcluster(
                linear_sum=unpack, mol_indices=[self.index_tracker]
            )
            split = self.root_.insert_bf_subcluster(subcluster, set_bits)

            if split:
                new_subcluster1, new_subcluster2 = _split_node(
                    self.root_, threshold, branching_factor
                )
                del self.root_
                self.root_ = _BFNode(
                    threshold=threshold,
                    branching_factor=branching_factor,
                    is_leaf=False,
                    n_features=n_features,
                    dtype=d_type,
                )
                self.root_.append_subcluster(new_subcluster1)
                self.root_.append_subcluster(new_subcluster2)

            self.index_tracker += 1
        if store_centroids:
            centroids = np.concatenate([leaf.centroids_ for leaf in self._get_leaves()])
            self.subcluster_centers_ = centroids
            self._n_features_out = self.subcluster_centers_.shape[0]

        self.first_call = False
        return self

    def fit_np(self, X):
        threshold = self.threshold
        branching_factor = self.branching_factor

        n_features = X.shape[1] - 1
        d_type = X.dtype

        # If partial_fit is called for the first time or fit is called, we
        # start a new tree.
        if self.first_call:
            # The first root is the leaf. Manipulate this object throughout.
            self.root_ = _BFNode(
                threshold=threshold,
                branching_factor=branching_factor,
                is_leaf=True,
                n_features=n_features,
                dtype=d_type,
            )

            # To enable getting back subclusters.
            self.dummy_leaf_ = _BFNode(
                threshold=threshold,
                branching_factor=branching_factor,
                is_leaf=True,
                n_features=n_features,
                dtype=d_type,
            )
            self.dummy_leaf_.next_leaf_ = self.root_
            self.root_.prev_leaf_ = self.dummy_leaf_

        # Cannot vectorize. Enough to convince to use cython.
        if not sparse.issparse(X):
            iter_func = iter
        else:
            iter_func = _iterate_sparse_X

        for sample in iter_func(X):
            sample_copy = sample.copy()
            subcluster = _BFSubcluster(
                linear_sum=sample_copy[:-1], mol_indices=[self.index_tracker]
            )
            n_samples = sample_copy[-1]
            if n_samples > 1:
                subcluster.n_samples_ = n_samples
                subcluster.centroid_ = calc_centroid(sample_copy[:-1], n_samples)
            set_bits = np.sum(subcluster.centroid_.astype("uint64"))
            split = self.root_.insert_bf_subcluster(subcluster, set_bits)

            if split:
                new_subcluster1, new_subcluster2 = _split_node(
                    self.root_, threshold, branching_factor
                )
                del self.root_
                self.root_ = _BFNode(
                    threshold=threshold,
                    branching_factor=branching_factor,
                    is_leaf=False,
                    n_features=n_features,
                    dtype=d_type,
                )
                self.root_.append_subcluster(new_subcluster1)
                self.root_.append_subcluster(new_subcluster2)

            self.index_tracker += 1

        self.first_call = False
        return self

    def fit_np_reinsert(self, X, reinsert_indices):
        threshold = self.threshold
        branching_factor = self.branching_factor

        n_features = X.shape[1] - 1
        d_type = X.dtype

        # If partial_fit is called for the first time or fit is called, we
        # start a new tree.
        if self.first_call:
            # The first root is the leaf. Manipulate this object throughout.
            self.root_ = _BFNode(
                threshold=threshold,
                branching_factor=branching_factor,
                is_leaf=True,
                n_features=n_features,
                dtype=d_type,
            )

            # To enable getting back subclusters.
            self.dummy_leaf_ = _BFNode(
                threshold=threshold,
                branching_factor=branching_factor,
                is_leaf=True,
                n_features=n_features,
                dtype=d_type,
            )
            self.dummy_leaf_.next_leaf_ = self.root_
            self.root_.prev_leaf_ = self.dummy_leaf_

        # Cannot vectorize. Enough to convince to use cython.
        if not sparse.issparse(X):
            iter_func = iter
        else:
            iter_func = _iterate_sparse_X

        for sample, mol_inds in zip(iter_func(X), reinsert_indices):
            sample_copy = sample.copy()
            subcluster = _BFSubcluster(
                linear_sum=sample_copy[:-1], mol_indices=mol_inds
            )
            n_samples = sample_copy[-1]
            if n_samples > 1:
                subcluster.n_samples_ = n_samples
                subcluster.centroid_ = calc_centroid(sample_copy[:-1], n_samples)
            set_bits = np.sum(subcluster.centroid_.astype("uint64"))
            split = self.root_.insert_bf_subcluster(subcluster, set_bits)

            if split:
                new_subcluster1, new_subcluster2 = _split_node(
                    self.root_, threshold, branching_factor
                )
                del self.root_
                self.root_ = _BFNode(
                    threshold=threshold,
                    branching_factor=branching_factor,
                    is_leaf=False,
                    n_features=n_features,
                    dtype=d_type,
                )
                self.root_.append_subcluster(new_subcluster1)
                self.root_.append_subcluster(new_subcluster2)

        self.first_call = False
        return self

    def fit_reinsert(
        self,
        X,
        reinsert_indices,
        store_centroids: bool = False,
        input_is_packed: bool = True,
        n_features: int | None = None,
    ):
        """X corresponds to only the molecules that will be reinserted into the tree
        reinsert indices are the indices of the molecules that will be reinserted into the tree
        """
        threshold = self.threshold
        branching_factor = self.branching_factor
        n_features = _validate_n_features(X, input_is_packed, n_features)
        d_type = X.dtype

        # If partial_fit is called for the first time or fit is called, we
        # start a new tree.
        if self.first_call:
            # The first root is the leaf. Manipulate this object throughout.
            self.root_ = _BFNode(
                threshold=threshold,
                branching_factor=branching_factor,
                is_leaf=True,
                n_features=n_features,
                dtype=d_type,
            )

            # To enable getting back subclusters.
            self.dummy_leaf_ = _BFNode(
                threshold=threshold,
                branching_factor=branching_factor,
                is_leaf=True,
                n_features=n_features,
                dtype=d_type,
            )
            self.dummy_leaf_.next_leaf_ = self.root_
            self.root_.prev_leaf_ = self.dummy_leaf_

        # Cannot vectorize. Enough to convince to use cython.
        if not sparse.issparse(X):
            iter_func = iter
        else:
            iter_func = _iterate_sparse_X

        for sample, mol_ind in zip(iter_func(X), reinsert_indices):
            unpack = _copy_or_unpack(sample, n_features, input_is_packed)
            set_bits = np.sum(unpack.astype("uint64"))
            subcluster = _BFSubcluster(linear_sum=unpack, mol_indices=[mol_ind])
            split = self.root_.insert_bf_subcluster(subcluster, set_bits)
            if split:
                new_subcluster1, new_subcluster2 = _split_node(
                    self.root_, threshold, branching_factor
                )
                del self.root_
                self.root_ = _BFNode(
                    threshold=threshold,
                    branching_factor=branching_factor,
                    is_leaf=False,
                    n_features=n_features,
                    dtype=d_type,
                )
                self.root_.append_subcluster(new_subcluster1)
                self.root_.append_subcluster(new_subcluster2)

        if store_centroids:
            centroids = np.concatenate([leaf.centroids_ for leaf in self._get_leaves()])
            self.subcluster_centers_ = centroids
            self._n_features_out = self.subcluster_centers_.shape[0]

        self.first_call = False
        return self

    def _get_leaves(self):
        """
        Retrieve the leaves of the BF Node.

        Returns
        -------
        leaves : list of shape (n_leaves,)
            List of the leaf nodes.
        """
        leaf_ptr = self.dummy_leaf_.next_leaf_
        leaves = []
        while leaf_ptr is not None:
            leaves.append(leaf_ptr)
            leaf_ptr = leaf_ptr.next_leaf_
        return leaves

    def get_centroids_mol_ids(self):
        """Method to return a dictionary containing the centroids and mol indices of the leaves"""
        if self.first_call:
            raise ValueError("The model has not been fitted yet.")

        centroids = []
        mol_ids = []
        for leaf in self._get_leaves():
            for subcluster in leaf.subclusters_:
                centroids.append(subcluster.centroid_)
                mol_ids.append(subcluster.mol_indices)

        dict_centroids_mol_ids = {"centroids": centroids, "mol_ids": mol_ids}

        return dict_centroids_mol_ids

    def get_centroids(self):
        """Method to return a list of Numpy arrays containing the centroids' fingerprints"""
        if self.first_call:
            raise ValueError("The model has not been fitted yet.")

        centroids = []
        for leaf in self._get_leaves():
            for subcluster in leaf.subclusters_:
                centroids.append(subcluster.centroid_)

        return centroids

    def get_cluster_mol_ids(self):
        """Method to return the indices of molecules in each cluster"""
        if self.first_call:
            raise ValueError("The model has not been fitted yet.")

        clusters_mol_id = []
        for leaf in self._get_leaves():
            for subcluster in leaf.subclusters_:
                clusters_mol_id.append(subcluster.mol_indices)

        # Sort the clusters by the number of samples in the cluster
        clusters_mol_id = sorted(clusters_mol_id, key=lambda x: len(x), reverse=True)

        return clusters_mol_id

    def _get_BFs(self):
        """Method to return the BitFeatures of the leaves"""
        if self.first_call:
            raise ValueError("The model has not been fitted yet.")

        BFs = []
        for leaf in self._get_leaves():
            for subcluster in leaf.subclusters_:
                BFs.append(subcluster)

        # Sort the BitFeatures by the number of samples in the cluster
        BFs = sorted(BFs, key=lambda x: x.n_samples_, reverse=True)

        return BFs

    def bf_to_np_refine(
        self,
        fps,
        initial_mol=0,
        input_is_packed: bool = True,
        n_features: int | None = None,
    ):
        """Method to prepare the BitFeatures of the largest cluster and the rest of the clusters"""
        if self.first_call:
            raise ValueError("The model has not been fitted yet.")
        n_features = _validate_n_features(fps, input_is_packed, n_features)
        BFs = self._get_BFs()
        big, rest = BFs[0], BFs[1:]

        fp_64 = []
        fp_32 = []
        fp_16 = []
        fp_8 = []

        mols_64 = []
        mols_32 = []
        mols_16 = []
        mols_8 = []

        for BF in rest:
            if BF.n_samples_ >= 4294967294:
                fp_64.append(BF.linear_sum_)
                fp_64.append(BF.n_samples_)
                mols_64.append(BF.mol_indices)
            elif BF.n_samples_ >= 65534:
                fp_32.append(BF.linear_sum_)
                fp_32.append(BF.n_samples_)
                mols_32.append(BF.mol_indices)
            elif BF.n_samples_ >= 254:
                fp_16.append(list(BF.linear_sum_))
                fp_16[-1].append(BF.n_samples_)
                mols_16.append(BF.mol_indices)
            else:
                fp_8.append(list(BF.linear_sum_))
                fp_8[-1].append(BF.n_samples_)
                mols_8.append(BF.mol_indices)

        for mol in big.mol_indices:
            if input_is_packed:
                fp_8.append(
                    list(unpack_fingerprints(fps[mol - initial_mol], n_features))
                )
            else:
                fp_8.append(list(fps[mol - initial_mol]))
            fp_8[-1].append(1)
            mols_8.append(BF.mol_indices)

        fps_bfs = []
        mols_bfs = []
        if len(fp_64) != 0:
            fp_64 = np.array(fp_64, dtype=np.uint64)
            fps_bfs.append(fp_64)
            mols_bfs.append(mols_64)
        if len(fp_32) != 0:
            fp_32 = np.array(fp_32, dtype=np.uint32)
            fps_bfs.append(fp_32)
            mols_bfs.append(mols_32)
        if len(fp_16) != 0:
            fp_16 = np.array(fp_16, dtype=np.uint16)
            fps_bfs.append(fp_16)
            mols_bfs.append(mols_16)
        if len(fp_8) != 0:
            fp_8 = np.array(fp_8, dtype=np.uint8)
            fps_bfs.append(fp_8)
            mols_bfs.append(mols_8)

        return fps_bfs, mols_bfs

    def bf_to_np(self):
        """Method to prepare the BitFeatures of the largest cluster and the rest of the clusters"""
        if self.first_call:
            raise ValueError("The model has not been fitted yet.")

        BFs = self._get_BFs()

        fp_64 = []
        fp_32 = []
        fp_16 = []
        fp_8 = []

        mols_64 = []
        mols_32 = []
        mols_16 = []
        mols_8 = []

        for BF in BFs:
            if BF.n_samples_ >= 4294967294:
                fp_64.append(BF.linear_sum_)
                fp_64.append(BF.n_samples_)
                mols_64.append(BF.mol_indices)
            elif BF.n_samples_ >= 65534:
                fp_32.append(BF.linear_sum_)
                fp_32.append(BF.n_samples_)
                mols_32.append(BF.mol_indices)
            elif BF.n_samples_ >= 254:
                fp_16.append(list(BF.linear_sum_))
                fp_16[-1].append(BF.n_samples_)
                mols_16.append(BF.mol_indices)
            else:
                fp_8.append(list(BF.linear_sum_))
                fp_8[-1].append(BF.n_samples_)
                mols_8.append(BF.mol_indices)

        fps_bfs = []
        mols_bfs = []
        if len(fp_64) != 0:
            fp_64 = np.array(fp_64, dtype=np.uint64)
            fps_bfs.append(fp_64)
            mols_bfs.append(mols_64)
        if len(fp_32) != 0:
            fp_32 = np.array(fp_32, dtype=np.uint32)
            fps_bfs.append(fp_32)
            mols_bfs.append(mols_32)
        if len(fp_16) != 0:
            fp_16 = np.array(fp_16, dtype=np.uint16)
            fps_bfs.append(fp_16)
            mols_bfs.append(mols_16)
        if len(fp_8) != 0:
            fp_8 = np.array(fp_8, dtype=np.uint8)
            fps_bfs.append(fp_8)
            mols_bfs.append(mols_8)

        return fps_bfs, mols_bfs

    def get_assignments(self, n_mols):
        clustered_ids = self.get_cluster_mol_ids()

        assignments = np.full(n_mols, -1, dtype=int)
        for i, cluster in enumerate(clustered_ids):
            assignments[cluster] = i + 1

        # Check that there are no unassigned molecules
        assert np.all(assignments != -1)

        return assignments



# type: ignore
import gc
import math
import time
import sys
import os
from pathlib import Path
import tempfile
import pytest
import bblean

try:
    import bblean._cpp_similarity as csim  # noqa

    CSIM_AVAIL = True
except ImportError:
    if os.getenv("BITBIRCH_CANT_SKIP_CPP_TESTS"):
        raise
    CSIM_AVAIL = False

try:
    from memray import Tracker, FileReader
except Exception:
    # Not available in Windows
    pass

from bblean.fingerprints import make_fake_fingerprints


# NOTE: The reproducibility for allocations for 25k fps is not very good
def test_memory_regression(subtests) -> None:
    if "memray" not in sys.modules:
        pytest.skip(
            "memory regression tests require memray, only avaliable in Linux and macOS"
        )
    all_max_allowed_bytes = [43_000_000, 63_000_000, 86_000_000, 105_000_000]
    all_fps_nums = [10_000, 15_000, 20_000, 25_000]
    # Around 41.9 MB should be allocated for these 100k fps
    # If memory usage scaled linearly it would take ~ 3.7 GiB for 1M molecules
    # Actual benchmarked allocation is 41_914_658 for
    # - py3.11
    # - numpy 2.3
    # - ubuntu-24.04
    # - GLIBC 2.39
    for fps_num, max_allowed_bytes in zip(all_fps_nums, all_max_allowed_bytes):
        with subtests.test(fps_num=fps_num):
            fps = make_fake_fingerprints(fps_num, seed=4068791011890883085)
            with tempfile.TemporaryDirectory() as d:
                tmp_dir = Path(d)
                tree = bblean.BitBirch(branching_factor=50, threshold=0.65)
                with Tracker(tmp_dir / "memray.bin"):
                    tree.fit(fps)
                reader = FileReader(tmp_dir / "memray.bin")
                total_alloc_bytes = sum(
                    record.size
                    for record in reader.get_high_watermark_allocation_records(
                        merge_threads=True
                    )
                )
            print(
                f"Fps num: {fps_num}, Total alloc MB: {total_alloc_bytes / 1024**2},"
                f" Tot. alloc bytes: {total_alloc_bytes:_} (max: {max_allowed_bytes:_})"
            )
            assert total_alloc_bytes < max_allowed_bytes


# NOTE: This test is pretty fragile and may fail if CI machines change
def test_speed_regression(subtests) -> None:
    all_fps_nums = [10_000, 15_000, 20_000]
    # Fitting 15_000 fps should take ~ 1.15-1.00s or less:
    # - py3.11
    # - numpy 2.3
    # - ubuntu-24.04
    # - GLIBC 2.39
    # - AMD Ryzen 5 7535HS
    # For this system the following values are sufficient:
    # all_max_allowed_ns = [1_200_000_000, 1_900_000_000, 2_500_000_000]
    # For the ubuntu-24.04 in gh CI the following are required:
    if CSIM_AVAIL:
        all_max_allowed_ns = [900_000_000, 1_500_000_000, 2_000_000_000]
    else:
        all_max_allowed_ns = [1_700_000_000, 2_600_000_000, 3_600_000_000]
    for fps_num, max_allowed_ns in zip(all_fps_nums, all_max_allowed_ns):
        with subtests.test(fps_num=fps_num):
            fps = make_fake_fingerprints(fps_num, seed=4068791011890883085)
            repeats = 3

            total_time_ns = 0
            tree = bblean.BitBirch(branching_factor=50, threshold=0.65)
            for _ in range(repeats):
                start_ns = time.process_time_ns()
                tree.fit(fps)
                total_time_ns += time.process_time_ns() - start_ns
                tree.reset()
                gc.collect()
            total_time_ns = math.ceil(total_time_ns / repeats)
            print(
                f"Fps num: {fps_num}, Total time ns: {total_time_ns:_} "
                f" (max: {max_allowed_ns:_})"
            )
            assert total_time_ns < max_allowed_ns



r"""SMILES manipulation"""

import typing as tp
from numpy.typing import NDArray
import numpy as np
from pathlib import Path

from bblean.utils import batched

__all__ = [
    "load_smiles",
    "calc_num_smiles",
    "iter_smiles_from_paths",
]

SmilesPaths = tp.Iterable[Path | str] | Path | str


def load_smiles(smiles_paths: SmilesPaths, max_num: int = -1) -> NDArray[np.str_]:
    r"""Simple utility to load smiles from a ``*.smi`` file"""
    smiles = []
    for i, smi in enumerate(iter_smiles_from_paths(smiles_paths)):
        if i == max_num:
            break
        smiles.append(smi)
    return np.asarray(smiles)


def calc_num_smiles(smiles_paths: SmilesPaths) -> int:
    r"""Get the total number of smiles in a sequene of paths"""
    return sum(1 for _ in iter_smiles_from_paths(smiles_paths))


# NOTE: replace_dummy is the procedure used in RDKit as of Dec 2024 for Synthon spaces
# synthons marked with [U], [Np], [Pu], [Am]. These need to be converted
def iter_smiles_from_paths(
    smiles_paths: SmilesPaths,
    tab_separated: bool = False,
    replace_dummy_atoms: bool = False,
) -> tp.Iterator[str]:
    r"""Iterate over smiles in a sequence of smiles paths

    If tab_separated = True the file is assumed to have the format
    <smiles><tab><field><tab><field>..., and only the smiles is returned
    """
    if isinstance(smiles_paths, (Path, str)):
        smiles_paths = [smiles_paths]
    for smi_path in smiles_paths:
        with open(smi_path, mode="rt", encoding="utf-8") as f:
            for smi in f:
                if tab_separated:
                    smi = smi.split("\t")[0]

                # Skip headers
                if smi.lower().strip() == "smiles":
                    continue

                # Replace 'dummy' atoms from synthon spaces
                if replace_dummy_atoms:
                    smi = (
                        smi.replace("[U]", "[1*]")
                        .replace("[Np]", "[2*]")
                        .replace("[Pu]", "[3*]")
                        .replace("[Am]", "[4*]")
                    )
                yield smi


def _iter_ranges_and_smiles_batches(
    smiles_paths: SmilesPaths,
    num_per_batch: int,
    tab_separated: bool = False,
    replace_dummy_atoms: bool = False,
    assume_paths: bool = True,
) -> tp.Iterable[tuple[tuple[int, int], tuple[str, ...]]]:
    if assume_paths:
        it = iter_smiles_from_paths(smiles_paths, tab_separated, replace_dummy_atoms)
    else:
        it = tp.cast(tp.Iterator[str], smiles_paths)
    start_idx = 0
    for batch in batched(it, num_per_batch):
        size = len(batch)
        end_idx = start_idx + size
        yield (start_idx, end_idx), batch
        start_idx = end_idx


def _iter_idxs_and_smiles_batches(
    smiles_paths: SmilesPaths,
    num_per_batch: int,
    tab_separated: bool = False,
    replace_dummy_atoms: bool = False,
) -> tp.Iterable[tuple[int, tuple[str, ...]]]:
    yield from enumerate(
        batched(
            iter_smiles_from_paths(smiles_paths, tab_separated, replace_dummy_atoms),
            num_per_batch,
        )
    )



r"""Analysis of clustering results"""

import warnings
from pathlib import Path
from collections import defaultdict
import dataclasses
import typing as tp
from functools import cached_property

import pandas as pd
import numpy as np
from numpy.typing import NDArray
from rdkit.Chem.Scaffolds import MurckoScaffold

from bblean._config import DEFAULTS
from bblean.similarity import jt_isim
from bblean.fingerprints import (
    fps_from_smiles,
    unpack_fingerprints,
    pack_fingerprints,
    _FingerprintFileSequence,
)

__all__ = [
    "scaffold_analysis",
    "cluster_analysis",
    "ScaffoldAnalysis",
    "ClusterAnalysis",
]


@dataclasses.dataclass
class ScaffoldAnalysis:
    r""":meta private:"""

    unique_num: int
    isim: float


class ClusterAnalysis:
    r""":meta private:"""

    def __init__(
        self,
        selected_cluster_sizes: list[int],
        all_cluster_sizes: list[int],
        df: pd.DataFrame,
        total_fps_num: int,
        selected_fps: NDArray[np.uint8] | None = None,
        fps_are_packed: bool = True,
        n_features: int | None = None,
        min_size: int | None = None,
    ) -> None:
        self.total_fps = total_fps_num
        self.stats = pd.Series(all_cluster_sizes).describe()
        self._all_cluster_sizes = all_cluster_sizes
        self._selected_cluster_sizes = selected_cluster_sizes
        self._fps = selected_fps
        self._df = df
        self.fps_are_packed = fps_are_packed
        self.n_features = n_features
        self.min_size = min_size

    def all_clusters_num_with_size_above(self, size: int) -> int:
        return sum(1 for c in self._all_cluster_sizes if c > size)

    @cached_property
    def all_singletons_num(self) -> int:
        return sum(1 for c in self._all_cluster_sizes if c == 1)

    def get_top_cluster_fps(self, packed: bool = True) -> list[NDArray[np.uint8]]:
        if self._fps is None:
            raise RuntimeError("Fingerprints not present")
        fps = self.top_packed_fps if packed else self.top_unpacked_fps
        out = []
        offset = 0
        for s in self._selected_cluster_sizes:
            out.append(fps[offset : offset + s])
            offset += s
        return out

    @property
    def all_clusters_mean_size(self) -> float:
        return float(self.stats["mean"])

    @property
    def all_clusters_median_size(self) -> int:
        return int(self.stats["50%"])

    @property
    def all_clusters_q1(self) -> int:
        return int(self.stats["25%"])

    @property
    def all_clusters_q3(self) -> int:
        return int(self.stats["75%"])

    @property
    def all_clusters_min_size(self) -> int:
        return int(self.stats["min"])

    @property
    def all_clusters_max_size(self) -> int:
        return int(self.stats["max"])

    @property
    def all_clusters_num(self) -> int:
        return int(self.stats["count"])

    @property
    def top_unpacked_fps(self) -> NDArray[np.uint8]:
        if self._fps is None:
            raise RuntimeError("Fingerprints not present")
        if self.fps_are_packed:
            return unpack_fingerprints(self._fps, self.n_features)
        return self._fps

    @property
    def top_packed_fps(self) -> NDArray[np.uint8]:
        if self._fps is None:
            raise RuntimeError("Fingerprints not present")
        if self.fps_are_packed:
            return self._fps
        return pack_fingerprints(self._fps)

    @property
    def has_scaffolds(self) -> bool:
        return "unique_scaffolds_num" in self._df.columns

    @property
    def has_fps(self) -> bool:
        return self._fps is not None

    @property
    def has_all_clusters(self) -> bool:
        return self.clusters_num == self.all_clusters_num

    @property
    def clusters_num(self) -> int:
        return len(self._df)

    @property
    def isims(self) -> pd.Series:
        return self._df["isim"]

    @property
    def has_isims(self) -> bool:
        return "isim" in self._df

    @property
    def labels(self) -> pd.Series:
        return self._df["labels"]

    @property
    def sizes(self) -> pd.Series:
        return self._df["sizes"]

    @property
    def unique_scaffolds_num(self) -> pd.Series:
        return self._df["unique_scaffolds_num"]

    @property
    def unique_scaffolds_isim(self) -> pd.Series:
        return self._df["unique_scaffolds_isim"]

    def dump_metrics(self, path: Path) -> None:
        self._df.to_csv(path, index=False)


# Get the number of unique scaffolds and the scaffold isim
def scaffold_analysis(
    smiles: tp.Iterable[str],
    fp_kind: str = DEFAULTS.fp_kind,
    warn_nan: bool = True,
) -> ScaffoldAnalysis:
    r"""Perform a scaffold analysis of a sequence of smiles

    Note that the order of the input smiles is not relevant
    """
    if isinstance(smiles, str):
        smiles = [smiles]
    scaffolds = [MurckoScaffold.MurckoScaffoldSmilesFromSmiles(smi) for smi in smiles]
    unique_scaffolds = set(scaffolds)
    scaffolds_fps = fps_from_smiles(unique_scaffolds, kind=fp_kind, pack=False)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "default" if warn_nan else "ignore",
            category=RuntimeWarning,
            message="Invalid n_objects.*",
        )
        scaffolds_isim = jt_isim(scaffolds_fps, input_is_packed=False)
    return ScaffoldAnalysis(len(unique_scaffolds), scaffolds_isim)


def cluster_analysis(
    clusters: list[list[int]],
    fps: NDArray[np.integer] | Path | tp.Sequence[Path] | None = None,
    smiles: tp.Iterable[str] = (),
    n_features: int | None = None,
    top: int | None = 20,
    assume_sorted: bool = True,
    scaffold_fp_kind: str = DEFAULTS.fp_kind,
    input_is_packed: bool = True,
    min_size: int = 0,
    warn_nan: bool = True,
) -> ClusterAnalysis:
    r"""Perform a cluster analysis starting from clusters, smiles, and fingerprints"""
    if isinstance(smiles, str):
        smiles = [smiles]
    smiles = np.asarray(smiles)

    if not assume_sorted:
        # Largest first
        clusters = sorted(clusters, key=lambda x: len(x), reverse=True)
    all_cluster_sizes = [len(c) for c in clusters]
    total_fps = sum(all_cluster_sizes)
    # Filter by min size
    _clusters = []
    for i, c in enumerate(clusters):
        if all_cluster_sizes[i] < min_size:
            break
        if top is not None and i >= top:
            break
        _clusters.append(c)
    clusters = _clusters

    info: dict[str, list[tp.Any]] = defaultdict(list)
    fps_provider: tp.Union[_FingerprintFileSequence, NDArray[np.uint8], None]
    if fps is None:
        fps_provider = None
    elif isinstance(fps, Path):
        fps_provider = np.load(fps, mmap_mode="r")
    elif not isinstance(fps, np.ndarray):
        fps_provider = _FingerprintFileSequence(fps)
    else:
        fps_provider = tp.cast(NDArray[np.uint8], fps.astype(np.uint8, copy=False))

    if fps_provider is None:
        selected = None
    else:
        selected = np.empty(
            (sum(len(c) for c in clusters), fps_provider.shape[1]), dtype=np.uint8
        )
    start = 0
    for i, c in enumerate(clusters, 1):
        size = len(c)
        # If a file sequence is passed, the cluster indices must be sorted.
        # the cluster analysis is idx-order-independent, so this is fine
        info["labels"].append(i)
        info["sizes"].append(size)
        if smiles.size:
            analysis = scaffold_analysis(
                smiles[c], fp_kind=scaffold_fp_kind, warn_nan=warn_nan
            )
            info["unique_scaffolds_num"].append(analysis.unique_num)
            info["unique_scaffolds_isim"].append(analysis.isim)
        if fps_provider is not None:
            assert selected is not None
            _fps = fps_provider[sorted(c)]
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "default" if warn_nan else "ignore",
                    category=RuntimeWarning,
                    message="Invalid n_objects.*",
                )
                info["isim"].append(
                    jt_isim(
                        _fps, input_is_packed=input_is_packed, n_features=n_features
                    )
                )
            selected[start : start + size] = _fps
        start += size
    return ClusterAnalysis(
        [len(c) for c in clusters],
        all_cluster_sizes,
        pd.DataFrame(info),
        selected_fps=selected,
        total_fps_num=total_fps,
        fps_are_packed=input_is_packed,
        n_features=n_features,
        min_size=min_size,
    )



r"""Fallback python implementation of molecular similarity calculators"""

import warnings

from numpy.typing import NDArray
import numpy as np

from bblean.utils import min_safe_uint
from bblean.fingerprints import unpack_fingerprints


def centroid_from_sum(
    linear_sum: NDArray[np.integer], n_samples: int, *, pack: bool = True
) -> NDArray[np.uint8]:
    r"""Calculates the majority vote centroid from a sum of fingerprint values

    The majority vote centroid is an good approximation of the Tanimoto centroid.

    Parameters
    ----------

    linear_sum : np.ndarray
                 Sum of the elements column-wise
    n_samples : int
                Number of samples
    pack : bool
        Whether to pack the resulting fingerprints

    Returns
    -------
    centroid : np.ndarray[np.uint8]
               Centroid fingerprints of the given set
    """
    # NOTE: Numpy guarantees bools are stored as 0xFF -> True and 0x00 -> False,
    # so this view is fully safe
    if n_samples <= 1:
        centroid = linear_sum.astype(np.uint8, copy=False)
    else:
        centroid = (linear_sum >= n_samples * 0.5).view(np.uint8)
    if pack:
        return np.packbits(centroid, axis=-1)
    return centroid


def centroid(
    fps: NDArray[np.uint8],
    input_is_packed: bool = True,
    n_features: int | None = None,
    *,
    pack: bool = True,
) -> NDArray[np.uint8]:
    r"""Calculates the majority vote centroid from a set of fingerprints

    The majority vote centroid is an good approximation of the Tanimoto centroid.
    """
    if input_is_packed:
        fps = unpack_fingerprints(fps, n_features)
    return centroid_from_sum(
        np.sum(fps, axis=0, dtype=np.uint64),  # type: ignore
        len(fps),
        pack=pack,
    )


def jt_compl_isim(
    fps: NDArray[np.uint8], input_is_packed: bool = True, n_features: int | None = None
) -> NDArray[np.float64]:
    """Get all complementary (Tanimoto) similarities of a set of fps, using iSIM"""
    if input_is_packed:
        fps = unpack_fingerprints(fps, n_features)
    # Vectorized calculation of all compl isim
    # For all compl isim N is N_total - 1
    n_objects = len(fps) - 1
    if n_objects < 2:
        msg = "Invalid fps. len(fps) must be >= 3"
        warnings.warn(msg, RuntimeWarning, stacklevel=2)
        return np.full(len(fps), fill_value=np.nan, dtype=np.float64)
    linear_sum = np.sum(fps, axis=0)
    comp_sims = [jt_isim_from_sum(linear_sum - fp, n_objects) for fp in fps]
    return np.array(comp_sims, dtype=np.float64)


# Requires numpy >= 2.0
def _popcount(a: NDArray[np.uint8]) -> NDArray[np.uint32]:
    # a is packed uint8 array with last axis = bytes
    # Sum bit-counts across bytes to get per-object totals

    # If the array has columns that are a multiple of 8, doing a bitwise count
    # over the buffer reinterpreted as uint64 is slightly faster.
    # This is zero cost if the exception is not triggered. Not having a be a multiple of
    # 8 is a very unlikely scenario, since fps are typically 1024 or 2048
    b: NDArray[np.integer]
    try:
        b = a.view(np.uint64)
    except ValueError:
        b = a
    return np.bitwise_count(b).sum(axis=-1, dtype=np.uint32)


# O(N) approximation to obtain "most dissimilar fingerprints" within an array
def jt_most_dissimilar_packed(
    Y: NDArray[np.uint8], n_features: int | None = None
) -> tuple[np.integer, np.integer, NDArray[np.float64], NDArray[np.float64]]:
    """Finds two fps in a packed fp array that are the most Tanimoto-dissimilar

    This is not guaranteed to find the most dissimilar fps, it is
    a robust O(N) approximation that doesn't affect final cluster quality.
    First find centroid of Y, then find fp_1, the most dissimilar molecule
    to the centroid. Finally find fp_2, the most dissimilar molecule to fp_1

    Returns
    -------
    fp_1 : int
        index of the first fingerprint
    fp_2 : int
        index of the second fingerprint
    sims_fp_1 : np.ndarray
        Tanimoto similarities of Y to fp_1
    sims_fp_2: np.ndarray
        Tanimoto similarities of Y to fp_2
    """
    # Get the centroid of the fps
    n_samples = len(Y)
    Y_unpacked = unpack_fingerprints(Y, n_features)
    # np.sum() automatically promotes to uint64 unless forced to a smaller dtype
    linear_sum = np.sum(Y_unpacked, axis=0, dtype=min_safe_uint(n_samples))
    packed_centroid = centroid_from_sum(linear_sum, n_samples, pack=True)

    cardinalities = _popcount(Y)

    # Get similarity of each fp to the centroid, and the least similar fp idx (fp_1)
    sims_cent = _jt_sim_packed_precalc_cardinalities(Y, packed_centroid, cardinalities)
    fp_1 = np.argmin(sims_cent)

    # Get similarity of each fp to fp_1, and the least similar fp idx (fp_2)
    sims_fp_1 = _jt_sim_packed_precalc_cardinalities(Y, Y[fp_1], cardinalities)
    fp_2 = np.argmin(sims_fp_1)

    # Get similarity of each fp to fp_2
    sims_fp_2 = _jt_sim_packed_precalc_cardinalities(Y, Y[fp_2], cardinalities)
    return fp_1, fp_2, sims_fp_1, sims_fp_2


def _jt_sim_arr_vec_packed(
    x: NDArray[np.uint8],
    y: NDArray[np.uint8],
) -> NDArray[np.float64]:
    r"""Tanimoto similarity between packed fingerprints

    Either both inputs are vectors of shape (F,) (Numpy scalar is returned), or one is
    an vector (F,) and the other an array of shape (N, F) (Numpy array of shape (N,) is
    returned).
    """
    if x.ndim != 2 or y.ndim != 1:
        raise ValueError("Expected a 2D array and a 1D vector as inputs")
    return _jt_sim_packed_precalc_cardinalities(x, y, _popcount(x))


def _jt_sim_packed_precalc_cardinalities(
    x: NDArray[np.uint8],
    y: NDArray[np.uint8],
    cardinalities: NDArray[np.integer],
) -> NDArray[np.float64]:
    # _cardinalities must be the result of calling _popcount(arr)

    # Maximum value in the denominator sum is the 2 * n_features (which is typically
    # uint16, but we use uint32 for safety)
    intersection = _popcount(np.bitwise_and(x, y))

    # Return value requires an out-of-place operation since it casts uints to f64
    #
    # There may be NaN in the similarity array if the both the cardinality
    # and the vector are just zeros, in which case the intersection is 0 -> 0 / 0
    #
    # In these cases the fps are equal so the similarity *should be 1*, so we
    # clamp the denominator, which is A | B (zero only if A & B is zero too).
    return intersection / np.maximum(cardinalities + _popcount(y) - intersection, 1)


def jt_isim_unpacked(arr: NDArray[np.integer]) -> float:
    # cast is slower
    return jt_isim_from_sum(
        np.sum(arr, axis=0, dtype=np.uint64), len(arr)  # type: ignore
    )


def jt_isim_packed(fps: NDArray[np.integer], n_features: int | None = None) -> float:
    # cast is slower
    return jt_isim_from_sum(
        np.sum(
            unpack_fingerprints(fps, n_features),  # type: ignore
            axis=0,
            dtype=np.uint64,
        ),
        len(fps),
    )


def jt_isim_from_sum(linear_sum: NDArray[np.integer], n_objects: int) -> float:
    r"""iSIM Tanimoto, from sum of rows of a fingerprint array and number of rows

    iSIM Tanimoto was first propsed in:
    https://pubs.rsc.org/en/content/articlelanding/2024/dd/d4dd00041b

    :math:`iSIM_{JT}(X)` is an excellent :math:`O(N)` approximation of the average
    Tanimoto similarity of a set of fingerprints.

    Also equivalent to the complement of the Tanimoto diameter
    :math:`iSIM_{JT}(X) = 1 - D_{JT}(X)`.

    Parameters
    ----------
    c_total : np.ndarray
              Sum of the elements from an array of fingerprints X, column-wise
              c_total = np.sum(X, axis=0)

    n_objects : int
                Number of elements
                n_objects = X.shape[0]

    Returns
    ----------
    isim : float
           iSIM Jaccard-Tanimoto value
    """
    if n_objects < 2:
        warnings.warn(
            f"Invalid n_objects = {n_objects} in isim. Expected n_objects >= 2",
            RuntimeWarning,
            stacklevel=2,
        )
        return np.nan

    x = linear_sum.astype(np.uint64, copy=False)
    sum_kq = np.sum(x)
    # isim of fingerprints that are all zeros should be 1 (they are all equal)
    if sum_kq == 0:
        return 1
    sum_kqsq = np.dot(x, x)  # *dot* conserves dtype
    a = (sum_kqsq - sum_kq) / 2  # 'a' is scalar f64
    return a / (a + n_objects * sum_kq - sum_kqsq)



"""Optimized molecular similarity calculators"""

import os
import warnings

from numpy.typing import NDArray
import numpy as np

# NOTE: The most expensive calculation is *jt_sim_packed*, followed by _popcount_2d,
# centroid_from_sum, packing and unpacking
# TODO: Packing and unpacking *should be done in C++ using a lookup table*
__all__ = [
    # JT sim between two (sets of) fingerprints, and average tanimoto (using iSIM)
    "jt_isim_from_sum",
    "jt_isim",
    "jt_sim_packed",
    "jt_most_dissimilar_packed",
    # Radius and diameter from sum
    "jt_isim_radius_from_sum",
    "jt_isim_radius_compl_from_sum",
    "jt_isim_diameter_from_sum",
    # Radius and diameter from fps (packed and unpacked)
    "jt_isim_radius",
    "jt_isim_radius_compl",
    "jt_isim_diameter",
    # Centroid and medoid
    # Radius and diameter unpacked / packed
    "centroid_from_sum",
    "centroid",
    "jt_isim_medoid",
    # Complementary similarity
    "jt_compl_isim",
    "jt_stratified_sampling",
    "jt_sim_matrix_packed",
]

from bblean._py_similarity import centroid_from_sum, centroid
from bblean.fingerprints import pack_fingerprints, unpack_fingerprints

# jt_isim_packed and jt_isim_unpacked are not exposed, only used within functions for
# speed

if os.getenv("BITBIRCH_NO_EXTENSIONS"):
    from bblean._py_similarity import (
        jt_isim_from_sum,
        jt_isim_unpacked,
        jt_isim_packed,
        jt_compl_isim,
        _jt_sim_arr_vec_packed,
        jt_most_dissimilar_packed,
    )
else:
    try:
        from bblean._cpp_similarity import (  # type: ignore
            jt_isim_from_sum,
            jt_isim_unpacked_u8,
            jt_isim_packed_u8,
            jt_compl_isim,  # TODO: Does it need wrappers for non-uint8?
            _jt_sim_arr_vec_packed,
            jt_most_dissimilar_packed,
            # Needed for wrappers
            unpack_fingerprints as _unpack_fingerprints,
        )

        # Wrap these two since doing
        def jt_isim_unpacked(arr: NDArray[np.integer]) -> float:
            # Wrapping like this is slightly faster than letting pybind11 autocast
            if arr.dtype == np.uint64:
                return jt_isim_from_sum(
                    np.sum(arr, axis=0, dtype=np.uint64), len(arr)  # type: ignore
                )
            return jt_isim_unpacked_u8(arr)

        # Probably a mypy bug
        def jt_isim_packed(  # type: ignore
            arr: NDArray[np.integer], n_features: int | None = None
        ) -> float:
            # Wrapping like this is slightly faster than letting pybind11 autocast
            if arr.dtype == np.uint64:
                return jt_isim_from_sum(
                    np.sum(
                        _unpack_fingerprints(arr, n_features),  # type: ignore
                        axis=0,
                        dtype=np.uint64,
                    ),
                    len(arr),
                )
            return jt_isim_packed_u8(arr)

    except ImportError:
        from bblean._py_similarity import (  # type: ignore
            jt_isim_from_sum,
            jt_isim_unpacked,
            jt_isim_packed,
            jt_compl_isim,
            _jt_sim_arr_vec_packed,
            jt_most_dissimilar_packed,
        )

        warnings.warn(
            "C++ optimized similarity calculations not available,"
            " falling back to python implementation"
        )


def jt_isim_medoid(
    fps: NDArray[np.uint8],
    input_is_packed: bool = True,
    n_features: int | None = None,
    pack: bool = True,
) -> tuple[int, NDArray[np.uint8]]:
    r"""Calculate the (Tanimoto) medoid of a set of fingerprints, using iSIM

    Returns both the index of the medoid in the input array and the medoid itself

    .. note::
        Returns the first (or only) fingerprint for array of size 2 and 1 respectively.
        Raises ValueError for arrays of size 0

    """
    if not fps.size:
        raise ValueError("Size of fingerprints set must be > 0")
    if input_is_packed:
        fps = unpack_fingerprints(fps, n_features)

    if len(fps) < 3:
        idx = 0  # Medoid undefined for sets of 3 or more fingerprints
    else:
        idx = np.argmin(jt_compl_isim(fps, input_is_packed=False)).item()
    m = fps[idx]
    if pack:
        return idx, pack_fingerprints(m)
    return idx, m


def jt_isim(
    fps: NDArray[np.integer],
    input_is_packed: bool = True,
    n_features: int | None = None,
) -> float:
    r"""Average Tanimoto, using iSIM

    iSIM Tanimoto was first propsed in:
    https://pubs.rsc.org/en/content/articlelanding/2024/dd/d4dd00041b

    :math:`iSIM_{JT}(X)` is an excellent :math:`O(N)` approximation of the average
    Tanimoto similarity of a set of fingerprints.

    Also equivalent to the complement of the Tanimoto diameter
    :math:`iSIM_{JT}(X) = 1 - D_{JT}(X)`.

    Parameters
    ----------
    arr : np.ndarray
        2D fingerprint array

    input_is_packed : bool
        Whether the input array has packed fingerprints

    n_features: int | None
        Number of features when unpacking fingerprints. Only required if
        not a multiple of 8

    Returns
    ----------
    isim : float
           iSIM Jaccard-Tanimoto value
    """
    if input_is_packed:
        return jt_isim_packed(fps, n_features)
    return jt_isim_unpacked(fps)


def jt_isim_diameter(
    arr: NDArray[np.integer],
    input_is_packed: bool = True,
    n_features: int | None = None,
) -> float:
    r"""Calculate the Tanimoto diameter of a set of fingerprints"""
    return jt_isim_diameter_from_sum(
        np.sum(
            (
                unpack_fingerprints(arr.astype(np.uint8, copy=False), n_features)
                if input_is_packed
                else arr
            ),
            axis=0,
            dtype=np.uint64,
        ),  # type: ignore
        len(arr),
    )


def jt_isim_radius(
    arr: NDArray[np.integer],
    input_is_packed: bool = True,
    n_features: int | None = None,
) -> float:
    r"""Calculate the Tanimoto radius of a set of fingerprints"""
    return jt_isim_radius_from_sum(
        np.sum(
            (
                unpack_fingerprints(arr.astype(np.uint8, copy=False), n_features)
                if input_is_packed
                else arr
            ),
            axis=0,
            dtype=np.uint64,
        ),  # type: ignore
        len(arr),
    )


def jt_isim_radius_compl(
    arr: NDArray[np.integer],
    input_is_packed: bool = True,
    n_features: int | None = None,
) -> float:
    r"""Calculate the complement of the Tanimoto radius of a set of fingerprints"""
    return jt_isim_radius_compl_from_sum(
        np.sum(
            (
                unpack_fingerprints(arr.astype(np.uint8, copy=False), n_features)
                if input_is_packed
                else arr
            ),
            axis=0,
            dtype=np.uint64,
        ),  # type: ignore
        len(arr),
    )


def jt_isim_radius_compl_from_sum(ls: NDArray[np.integer], n: int) -> float:
    r"""Calculate the complement of the Tanimoto radius of a set of fingerprints"""
    #  Calculates 1 - R = Rc
    # NOTE: Use uint64 sum since jt_isim_from_sum casts to uint64 internally
    # This prevents multiple casts
    new_unpacked_centroid = centroid_from_sum(ls, n, pack=False)
    new_ls_1 = np.add(ls, new_unpacked_centroid, dtype=np.uint64)
    new_n_1 = n + 1
    new_jt = jt_isim_from_sum(ls, n)
    new_jt_1 = jt_isim_from_sum(new_ls_1, new_n_1)
    return (new_jt_1 * new_n_1 - new_jt * (n - 1)) / 2


def jt_isim_radius_from_sum(ls: NDArray[np.integer], n: int) -> float:
    r"""Calculate the Tanimoto radius of a set of fingerprints"""
    return 1 - jt_isim_radius_compl_from_sum(ls, n)


def jt_isim_diameter_from_sum(ls: NDArray[np.integer], n: int) -> float:
    r"""Calculate the Tanimoto diameter of a set of fingerprints.

    Equivalent to ``1 - jt_isim_from_sum(ls, n)``"""
    return 1 - jt_isim_from_sum(ls, n)


# General wrapper that works both in C++ and python
def jt_sim_packed(
    x: NDArray[np.uint8],
    y: NDArray[np.uint8],
) -> NDArray[np.float64]:
    r"""Tanimoto similarity between packed fingerprints

    Either both inputs are vectors of shape (F,) (Numpy scalar is returned), or
    one is an vector (F,) and the other an array of shape (N, F) (Numpy array of
    shape (N,) is returned).
    """
    if x.ndim == 1 and y.ndim == 1:
        return _jt_sim_arr_vec_packed(x.reshape(1, -1), y)[0]
    if x.ndim == 2:
        return _jt_sim_arr_vec_packed(x, y)
    if y.ndim == 2:
        return _jt_sim_arr_vec_packed(y, x)
    raise ValueError(
        "Expected either two 1D vectors, or one 1D vector and one 2D array"
    )


def jt_sim_matrix_packed(arr: NDArray[np.uint8]) -> NDArray[np.float64]:
    r"""Tanimoto similarity matrix between all pairs of packed fps in arr"""
    matrix = np.ones((len(arr), len(arr)), dtype=np.float64)
    for i in range(len(arr)):
        # Set the similarities for each row
        matrix[i, i + 1 :] = jt_sim_packed(arr[i], arr[i + 1 :])
        # Set the similarities for each column (symmetric)
        matrix[i + 1 :, i] = matrix[i, i + 1 :]
    return matrix


def estimate_jt_std(
    fps: NDArray[np.uint8],
    n_samples: int | None = None,
    input_is_packed: bool = True,
    n_features: int | None = None,
    max_samples: int = 1_000_000,
) -> float:
    r"""Estimate the std of all pairwise Tanimoto.

    Returns
    -------
    std : float
        The standard deviation of all pairwise Tanimoto among the sampled fingerprints.
    """
    num_fps = len(fps)
    if num_fps > max_samples:
        rng = np.random.default_rng(42)
        random_choices = rng.choice(num_fps, size=max_samples, replace=False)
        fps = fps[random_choices]
        num_fps = len(fps)
    if n_samples is None:
        # Heuristic: use at least 50 samples, or 1 per 10,000 fingerprints,
        # to balance statistical representativeness and computational efficiency
        # TODO: This heuristic is broken, too few samples until 500k
        if num_fps <= 500_000:
            n_samples = 50
        else:
            n_samples = num_fps // 10_000
    sample_idxs = jt_stratified_sampling(fps, n_samples, input_is_packed, n_features)

    # Work with only the sampled fingerprints
    fps = fps[sample_idxs]
    num_fps = len(fps)
    pairs = np.empty(num_fps * (num_fps - 1) // 2, dtype=np.float64)
    # NOTE: Calc upper triangular part of pairwise matrix only, slightly more efficient,
    # but difference is negligible in tests
    offset = 0
    for i in range(len(fps)):
        num = num_fps - i - 1
        pairs[offset : offset + num] = jt_sim_packed(fps[i], fps[i + 1 :])
        offset += num
    return np.std(pairs).item()


def jt_stratified_sampling(
    fps: NDArray[np.uint8],
    n_samples: int,
    input_is_packed: bool = True,
    n_features: int | None = None,
) -> NDArray[np.int64]:
    r"""Sample from a set of fingerprints according to their complementary similarity

    Given a group of fingerprints, calculate all complementary similarities, order, and
    sample the first element from consecutive groups of length ``num_fps // n_samples +
    1``.

    ..  note ::

        This is not true statistical stratified sampling, it is not random, and the
        strata are not homogeneous. It is meant as a reliable, deterministic method to
        obtain a representative sample from a set of fingerprints.
    """
    # Stratified sampling without replacement
    if n_samples == 0:
        return np.array([], dtype=np.int64)
    if n_samples > len(fps):
        raise ValueError("n_samples must be <= len(fps)")
    # Get the indices that would sort the complementary similarities
    sorted_indices = np.argsort(jt_compl_isim(fps, input_is_packed, n_features))
    # Split into n_samples strata
    strata = np.array_split(sorted_indices, n_samples)
    # Get first index of each strata
    return np.array([s[0] for s in strata])



import numpy as np


from bblean._legacy.bb_int64 import (  # type: ignore
    BitBirch as BitBirchInt64,
    set_merge as set_merge_int64,
)
from bblean._legacy.bb_uint8 import (  # type: ignore
    BitBirch as BitBirchUint8,
    set_merge as set_merge_uint8,
)
from bblean.bitbirch import BitBirch  # type: ignore
from bblean.fingerprints import make_fake_fingerprints, unpack_fingerprints


def test_random_fps_consistency() -> None:
    fps = make_fake_fingerprints(
        3000, n_features=2048, seed=12620509540149709235, pack=True
    )
    ids_expect = [
        [2195, 2196, 2378, 2440, 2443, 2454, 2463, 2464, 2465, 2467, 2527, 2544],
        [199, 228, 255, 270, 273, 438, 457, 458, 461, 470, 477, 496],
        [700, 728, 773, 798, 825, 891, 919, 962, 963, 968, 998],
        [1448, 1567, 1590, 1606, 1612, 1637, 1640, 1648, 1686, 1694],
        [1059, 1065, 1072, 1077, 1154, 1194, 1301],
        [1779, 1802, 1807, 1828, 1856, 1864],
        [2826, 2896, 2970, 2973, 2975],
        [1986, 2107, 2139, 2141],
        [1933, 1949],
        [2233, 2294],
        [1551, 1552],
        [1219, 1226],
        [614, 637],
    ]
    tree = BitBirch(branching_factor=50, threshold=0.65, merge_criterion="diameter")
    tree.fit(fps, n_features=2048)
    output = tree.get_cluster_mol_ids()
    assert output[:13] == ids_expect

    # Check consistency with "upacked" bitbirch
    fps = unpack_fingerprints(fps, n_features=2048)
    tree = BitBirch(branching_factor=50, threshold=0.65, merge_criterion="diameter")
    tree.fit(fps, n_features=2048, input_is_packed=False)
    output = tree.get_cluster_mol_ids()
    assert output[:13] == ids_expect

    # Check consistency with "legacy" bitbirchs (uint8 and int64)
    set_merge_uint8("diameter")
    tree = BitBirchUint8(threshold=0.65, branching_factor=50)
    tree.fit(fps, n_features=2048, input_is_packed=False)
    output = tree.get_cluster_mol_ids()
    assert output[:13] == ids_expect

    fps = fps.astype(np.int64)
    set_merge_int64("diameter")
    tree = BitBirchInt64(threshold=0.65, branching_factor=50)
    tree.fit(fps, n_features=2048, input_is_packed=False)
    output = tree.get_cluster_mol_ids()
    assert output[:13] == ids_expect


def test_random_fps_radius_consistency() -> None:
    fps = make_fake_fingerprints(
        1000, n_features=2048, seed=12620509540149709235, pack=True
    )
    tree = BitBirch(merge_criterion="radius", threshold=0.65, branching_factor=50)
    tree.fit(fps, n_features=2048)
    ids_expect = [
        [749, 751, 766, 798, 825, 827, 840, 845],
        [463, 464, 470, 477, 496, 498],
        [0, 3, 32, 53],
        [607, 609, 614, 615],
        [542, 557, 560, 561],
        [30, 36, 45],
        [647, 650, 652],
        [689, 694, 745],
        [762, 764, 773],
        [771, 775, 789],
        [520, 522, 551],
        [60, 66, 99],
        [248, 389, 390],
        [336, 391, 405],
        [199, 230, 231],
        [71, 125],
        [82, 155],
    ]
    output = tree.get_cluster_mol_ids()
    assert output[2:19] == ids_expect

    # Check consistency with "upacked" bitbirch
    fps = unpack_fingerprints(fps, n_features=2048)
    tree = BitBirch(merge_criterion="radius", threshold=0.65, branching_factor=50)
    tree.fit(fps, n_features=2048, input_is_packed=False)
    output = tree.get_cluster_mol_ids()
    assert output[2:19] == ids_expect

    # Check consistency with "legacy" bitbirchs (uint8 and int64)
    set_merge_uint8("radius")
    tree = BitBirchUint8(threshold=0.65, branching_factor=50)
    tree.fit(fps, n_features=2048, input_is_packed=False)
    output = tree.get_cluster_mol_ids()
    assert output[2:19] == ids_expect

    fps = fps.astype(np.int64)
    set_merge_int64("radius")
    tree = BitBirchInt64(threshold=0.65, branching_factor=50)
    tree.fit(fps, n_features=2048, input_is_packed=False)
    output = tree.get_cluster_mol_ids()
    assert output[2:19] == ids_expect


def test_random_fps_tolerance_consistency() -> None:
    fps = make_fake_fingerprints(
        500, n_features=2048, seed=12620509540149709235, pack=True
    )
    tree = BitBirch(
        merge_criterion="tolerance-legacy",
        branching_factor=50,
        threshold=0.65,
        tolerance=0.05,
    )
    tree.fit(fps, n_features=2048)
    output = tree.get_cluster_mol_ids()
    ids_expect = [[182, 255, 311, 389, 405, 438, 457, 461, 470], [107, 228], [13], [0]]
    assert output[:4] == ids_expect

    # Check consistency with "upacked" bitbirch
    fps = unpack_fingerprints(fps, n_features=2048)
    tree = BitBirch(
        merge_criterion="tolerance-legacy",
        branching_factor=50,
        threshold=0.65,
        tolerance=0.05,
    )
    tree.fit(fps, n_features=2048, input_is_packed=False)
    output = tree.get_cluster_mol_ids()
    assert output[:4] == ids_expect

    # Check consistency with "legacy" bitbirchs (uint8 and int64)
    set_merge_uint8("tolerance-legacy", 0.05)
    tree = BitBirchUint8(threshold=0.65, branching_factor=50)
    tree.fit(fps, n_features=2048, input_is_packed=False)
    output = tree.get_cluster_mol_ids()
    assert output[:4] == ids_expect

    fps = fps.astype(np.int64)
    set_merge_int64("tolerance-legacy", 0.05)
    tree = BitBirchInt64(threshold=0.65, branching_factor=50)
    tree.fit(fps, n_features=2048, input_is_packed=False)
    output = tree.get_cluster_mol_ids()
    assert output[:4] == ids_expect



r"""General timing tools"""

import json
import time
import typing as tp
from pathlib import Path

from rich.console import Console


class Timer:
    def __init__(self, indent: bool = True, log_endpoints: bool = False) -> None:
        self._timings_s: dict[str, float] = {}
        self._log_endpoints = log_endpoints
        self._start_timings_s: dict[str, float] = {}
        self._end_timings_s: dict[str, float] = {}
        self._indent = indent
        self._global_init_time_s: float | None = None
        self._global_end_time_s: float | None = None

    @property
    def timings_s(self) -> dict[str, float]:
        return self._timings_s.copy()

    def init_timing(self, label: str = "total") -> None:
        if label != "total" and self._global_init_time_s is None:
            raise ValueError("Overal timing needs to be initialized")

        if label in self._timings_s:
            raise ValueError(f"{label} has already been tracked")
        self._timings_s[label] = time.perf_counter()

        if label == "total":
            self._global_init_time_s = self._timings_s[label]

        if self._log_endpoints and label != "total":
            self._start_timings_s[label] = self._timings_s[label] - tp.cast(
                float, self._global_init_time_s
            )

    def end_timing(
        self,
        label: str = "total",
        console: Console | None = None,
        indent: bool | None = None,
    ) -> None:
        if self._global_end_time_s is not None:
            raise ValueError("Timing has finished, reset timing before reusing timer")

        indent = indent if indent is not None else self._indent
        if label not in self._timings_s:
            raise ValueError(f"{label} has not been initialized")
        end_time_s = time.perf_counter()
        self._timings_s[label] = end_time_s - self._timings_s[label]
        t = self._timings_s[label]

        if self._log_endpoints and label != "total":
            self._end_timings_s[label] = end_time_s - tp.cast(
                float, self._global_init_time_s
            )
        if console is not None:
            if indent:
                indent_str = "    "
            else:
                indent_str = ""
            if label == "total":
                console.print(f"{indent_str}- Total time elapsed: {t:.4f} s")
            else:
                console.print(f"{indent_str}- Time for {label}: {t:.4f} s")
        if label == "total":
            self._global_end_time_s = end_time_s

    def dump(self, path: Path) -> None:
        if self._global_end_time_s is None:
            raise ValueError("Timing not finished yet")
        output: dict[str, float | dict[str, float]] = {}
        output.update(self._timings_s)
        if self._start_timings_s:
            output["start_timings"] = self._start_timings_s
        if self._end_timings_s:
            output["end_timings"] = self._end_timings_s

        with open(path, mode="wt", encoding="utf-8") as f:
            json.dump(output, f, indent=4)

    def reset(self) -> None:
        self._timings_s = {}
        self._start_timings_s = {}
        self._end_timings_s = {}
        self._global_init_time_s = None
        self._global_end_time_s = None



r"""BitBIRCH-Lean, a high-throughput, memory-efficient implementation of BitBIRCH

BitBIRCH-Lean is designed for high-thorouput clustering of huge molecular
libraries (of up to hundreds of milliones of molecules).
"""

from bblean.smiles import load_smiles
from bblean.fingerprints import fps_from_smiles
from bblean.bitbirch import BitBirch, set_merge
from bblean.fingerprints import pack_fingerprints, unpack_fingerprints
from bblean._version import __version__

__all__ = [
    # Global namespace for convenience
    "BitBirch",
    "set_merge",
    "pack_fingerprints",
    "unpack_fingerprints",
    "load_smiles",
    "fps_from_smiles",
    "__version__",
]



# type: ignore
import bblean
import numpy as np
import pandas as pd
import bblean.similarity as iSIM


def recluster_iteration_metrics(
    fps: np.ndarray,
    threshold: float = None,
    sigmas: float = 3.5,
    extra: float = 0,
    branching_factor: int = 1024,
    iterations: int = 11,
) -> pd.DataFrame:

    if threshold is None:
        # Obtain the iSIM and iSIM-sigma
        isim = iSIM.jt_isim_packed(fps)
        isim_sigma = iSIM.estimate_jt_std(fps, n_samples=50)
        threshold = isim + sigmas * isim_sigma
    else:
        isim_sigma = iSIM.estimate_jt_std(fps, n_samples=50)

    # Do the initial clustering
    bbc = bblean.BitBirch(
        threshold=threshold,
        branching_factor=branching_factor,
        merge_criterion="diameter",
    )
    bbc.fit(fps)
    clusters = bbc.get_cluster_mol_ids()

    # Collect the initial statistics
    n_clusters = [len(clusters)]
    n_singletons = [len([c for c in clusters if len(c) == 1])]
    medoids = bbc.get_medoids(fps=fps)
    dispersion = [iSIM.jt_isim(medoids)]
    dispersion_no_singletons = [
        iSIM.jt_isim(medoids[: n_clusters[0] - n_singletons[0]])
    ]
    top_sizes = [[len(c) for c in clusters[:10]]]

    # Do refinement steps
    for _ in range(iterations):
        bbc.recluster_inplace(iterations=1, extra_threshold=extra * isim_sigma)
        clusters = bbc.get_cluster_mol_ids()

        # Collect statistics
        n_clusters.append(len(clusters))
        n_singletons.append(len([c for c in clusters if len(c) == 1]))
        medoids = bbc.get_medoids(fps=fps)
        dispersion.append(iSIM.jt_isim(medoids))
        end_idx = n_clusters[-1] - n_singletons[-1]
        dispersion_no_singletons.append(iSIM.jt_isim(medoids[:end_idx]))
        top_sizes.append([len(c) for c in clusters[:10]])

        if bbc.threshold >= 1.0:
            break

    # Save the results
    results = {
        "iteration": list(range(len(n_clusters))),
        "n_clusters": n_clusters,
        "n_singletons": n_singletons,
        "dispersions": dispersion,
        "dispersions_no_singletons": dispersion_no_singletons,
        "top_sizes": top_sizes,
    }

    df = pd.DataFrame(results)

    return df


def threshold_scan(
    fps: np.ndarray,
    max_sigmas: float = 10,
    branching_factor: int = 1024,
):

    # Obtain the iSIM and iSIM-sigma
    isim = iSIM.jt_isim_packed(fps)
    isim_sigma = iSIM.estimate_jt_std(fps, n_samples=50)

    # Do the clustering for different thresholds
    thresholds = []
    n_clusters = []
    n_singletons = []
    dispersions = []
    dispersions_non_singleton = []
    top_sizes = []
    top_range = min(isim + max_sigmas * isim_sigma, 1.0)
    for threshold in np.arange(isim, top_range, isim_sigma):
        bbc = bblean.BitBirch(
            threshold=threshold,
            branching_factor=branching_factor,
            merge_criterion="diameter",
        )
        bbc.fit(fps)
        clusters = bbc.get_cluster_mol_ids()

        medoids = bbc.get_medoids(fps=fps)

        # Collect statistics
        thresholds.append(threshold)
        n_clusters.append(len(clusters))
        n_singletons.append(len([c for c in clusters if len(c) == 1]))
        dispersions.append(iSIM.jt_isim(medoids))
        end_idx = n_clusters[-1] - n_singletons[-1]
        dispersions_non_singleton.append(iSIM.jt_isim(medoids[:end_idx]))
        top_sizes.append([len(c) for c in clusters[:10]])

        # Save the results
        results = {
            "thresholds": thresholds,
            "n_clusters": n_clusters,
            "n_singletons": n_singletons,
            "dispersions": dispersions,
            "dispersions_non_singleton": dispersions_non_singleton,
            "top_sizes": top_sizes,
        }

        df = pd.DataFrame(results)

    return df


def branching_factor_scan(
    fps: np.ndarray,
    sigmas_list: list[float],
    branching_factor_list: list[int] = [64, 128, 256, 512, 1024],
    save_path: str = None,
) -> pd.DataFrame:
    # Obtain the iSIM and iSIM-sigma
    isim = iSIM.jt_isim_packed(fps)
    isim_sigma = iSIM.estimate_jt_std(fps, n_samples=50)

    # Do the clustering for different thresholds
    thresholds_list = [isim + k * isim_sigma for k in sigmas_list]
    thresholds = []
    n_clusters = []
    n_singletons = []
    dispersions = []
    dispersions_non_singleton = []
    top_sizes = []
    branching_factors = []
    times = []
    memory = []
    for threshold in thresholds_list:
        for bf in branching_factor_list:
            bbc = bblean.BitBirch(
                threshold=threshold, branching_factor=bf, merge_criterion="diameter"
            )
            bbc.fit(fps)

            clusters = bbc.get_cluster_mol_ids()

            medoids = bbc.get_medoids(fps=fps)

            # Collect statistics
            thresholds.append(threshold)
            n_clusters.append(len(clusters))
            n_singletons.append(len([c for c in clusters if len(c) == 1]))
            branching_factors.append(bf)
            dispersions.append(iSIM.jt_isim(medoids))
            end_idx = n_clusters[-1] - n_singletons[-1]
            dispersions_non_singleton.append(iSIM.jt_isim(medoids[:end_idx]))
            top_sizes.append([len(c) for c in clusters[:10]])

            # Save the results
            results = {
                "thresholds": thresholds,
                "n_clusters": n_clusters,
                "n_singletons": n_singletons,
                "dispersions": dispersions,
                "dispersions_non_singleton": dispersions_non_singleton,
                "top_sizes": top_sizes,
                "branching_factor": branching_factors,
                "time": times,
                "memory_gb": memory,
            }

    df = pd.DataFrame(results)
    if save_path is not None:
        df.to_csv(save_path, index=False)
    else:
        return df



from pathlib import Path
import tempfile
import numpy as np
import itertools
from bblean.fingerprints import (
    _get_fingerprints_from_file_seq,
    make_fake_fingerprints,
    fps_from_smiles,
)
from bblean.smiles import load_smiles


def test_fps_from_smiles() -> None:
    smiles = load_smiles(Path(__file__).parent / "chembl-sample-3k.smi")
    fps_raw = fps_from_smiles(smiles).reshape(-1)
    nonzero = fps_raw.nonzero()[0].reshape(-1)
    actual = fps_raw[nonzero][:19].tolist()
    expect = [4, 128, 2, 16, 8, 16, 4, 16, 128, 16, 1, 128, 1, 64, 1, 1, 128, 32, 32]
    assert actual == expect


# NOTE: This is an acceptance test only
def test_fps_from_smiles_invalid() -> None:
    smiles = load_smiles(Path(__file__).parent / "chembl-sample-bad.smi")
    fps_raw, _ = fps_from_smiles(smiles, skip_invalid=True, sanitize="minimal")
    fps_raw = fps_raw.reshape(-1)
    nonzero = fps_raw.nonzero()[0].reshape(-1)
    actual = fps_raw[nonzero][:19].tolist()
    expect = [2, 4, 32, 1, 2, 128, 4, 128, 32, 32, 80, 128, 64, 128, 1, 16, 64, 4, 16]
    assert actual == expect


def test_fps_from_smiles_invalid_ids() -> None:
    smiles = load_smiles(Path(__file__).parent / "chembl-sample-bad.smi")
    fps_raw, invalid_idxs = fps_from_smiles(
        smiles, skip_invalid=True, sanitize="all"
    )
    actual = invalid_idxs.tolist()
    expect = [0, 1]
    assert actual == expect
    assert fps_raw.shape[0] == len(smiles) - len(expect)


def test_fingerprints_from_file_seq_empty() -> None:
    fps = make_fake_fingerprints(
        100, n_features=32, seed=12620509540149709235, pack=True
    )
    sections = [0, 10, 20, 30, 50, 80, 100]
    with tempfile.TemporaryDirectory() as d:
        paths = []
        for i, (start, end) in enumerate(itertools.pairwise(sections)):
            path = Path(d).resolve() / f"fps.{str(i).zfill(3)}.npy"
            paths.append(path)
            np.save(path, fps[start:end])

        idxs: list[int] = []
        expect = fps[idxs]
        actual = _get_fingerprints_from_file_seq(paths, idxs)
        assert (expect == actual).all()


def test_fingerprints_from_file_seq() -> None:
    fps = make_fake_fingerprints(
        100, n_features=32, seed=12620509540149709235, pack=True
    )
    sections = [0, 10, 20, 30, 50, 80, 100]
    with tempfile.TemporaryDirectory() as d:
        paths = []
        for i, (start, end) in enumerate(itertools.pairwise(sections)):
            path = Path(d).resolve() / f"fps.{str(i).zfill(3)}.npy"
            paths.append(path)
            np.save(path, fps[start:end])

        idxs = [0, 1, 2, 3, 4, 21, 22, 29, 50, 51, 55, 83]
        expect = fps[idxs]
        actual = _get_fingerprints_from_file_seq(paths, idxs)
        assert (expect == actual).all()


def test_fingerprints_from_file_seq_variation() -> None:
    fps = make_fake_fingerprints(
        100, n_features=32, seed=12620509540149709235, pack=True
    )
    sections = [0, 30, 80, 100]
    with tempfile.TemporaryDirectory() as d:
        paths = []
        for i, (start, end) in enumerate(itertools.pairwise(sections)):
            path = Path(d).resolve() / f"fps.{str(i).zfill(3)}.npy"
            paths.append(path)
            np.save(path, fps[start:end])

        idxs = list(range(100))
        expect = fps[idxs]
        actual = _get_fingerprints_from_file_seq(paths, idxs)
        assert (expect == actual).all()



import pickle
import numpy as np
from pathlib import Path
import tempfile
from typer.testing import CliRunner

from bblean.cli import app
from bblean.fingerprints import make_fake_fingerprints


def test_umap() -> None:
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as d:
        Path
        dir = Path(d).resolve()
        fps = make_fake_fingerprints(
            50, n_features=512, seed=12620509540149709235, pack=True
        )
        np.save(dir / "fingerprints.npy", fps)
        out_dir = dir / "output"
        result = runner.invoke(
            app,
            [
                "run",
                str(dir),
                "-o",
                str(out_dir),
                "-b",
                "50",
                "-t",
                "0.3",
                "--no-monitor-mem",
            ],
        )
        assert result.exit_code == 0
        result = runner.invoke(
            app,
            [
                "plot-umap",
                str(out_dir),
                "-f",
                str(dir / "fingerprints.npy"),
                "--no-show",
                "--top",
                "1",
                "--no-verbose",
            ],
        )
    assert result.exit_code == 0


def test_pops() -> None:
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as d:
        Path
        dir = Path(d).resolve()
        fps = make_fake_fingerprints(
            250, n_features=512, seed=12620509540149709235, pack=True
        )
        np.save(dir / "fingerprints.npy", fps)
        out_dir = dir / "output"
        result = runner.invoke(
            app,
            [
                "run",
                str(dir),
                "-o",
                str(out_dir),
                "-b",
                "50",
                "-t",
                "0.3",
                "--no-monitor-mem",
            ],
        )
        assert result.exit_code == 0
        result = runner.invoke(
            app,
            [
                "plot-pops",
                str(out_dir),
                "-f",
                str(dir / "fingerprints.npy"),
                "--min-size",
                "1",
                "--top",
                "2",
                "--no-show",
            ],
        )
        assert result.exit_code == 0


def test_pca() -> None:
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as d:
        Path
        dir = Path(d).resolve()
        fps = make_fake_fingerprints(
            250, n_features=512, seed=12620509540149709235, pack=True
        )
        np.save(dir / "fingerprints.npy", fps)
        out_dir = dir / "output"
        result = runner.invoke(
            app,
            [
                "run",
                str(dir),
                "-o",
                str(out_dir),
                "-b",
                "50",
                "-t",
                "0.3",
                "--no-monitor-mem",
            ],
        )
        assert result.exit_code == 0
        result = runner.invoke(
            app,
            [
                "plot-pca",
                str(out_dir),
                "-f",
                str(dir / "fingerprints.npy"),
                "--no-show",
                "--top",
                "5",
                "--no-verbose",
            ],
        )
    assert result.exit_code == 0


def test_tsne() -> None:
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as d:
        Path
        dir = Path(d).resolve()
        fps = make_fake_fingerprints(
            250, n_features=512, seed=12620509540149709235, pack=True
        )
        np.save(dir / "fingerprints.npy", fps)
        out_dir = dir / "output"
        result = runner.invoke(
            app,
            [
                "run",
                str(dir),
                "-o",
                str(out_dir),
                "-b",
                "50",
                "-t",
                "0.3",
                "--no-monitor-mem",
            ],
        )
        assert result.exit_code == 0
        result = runner.invoke(
            app,
            [
                "plot-tsne",
                str(out_dir),
                "-f",
                str(dir / "fingerprints.npy"),
                "--no-show",
                "--top",
                "2",
                "--no-verbose",
            ],
        )
    assert result.exit_code == 0


def test_summary() -> None:
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as d:
        Path
        dir = Path(d).resolve()
        fps = make_fake_fingerprints(
            500, n_features=2048, seed=12620509540149709235, pack=True
        )
        np.save(dir / "fingerprints.npy", fps)
        out_dir = dir / "output"
        result = runner.invoke(
            app,
            [
                "run",
                str(dir),
                "-o",
                str(out_dir),
                "-b",
                "50",
                "-t",
                "0.1",
                "--no-monitor-mem",
            ],
        )
        assert result.exit_code == 0
        result = runner.invoke(
            app,
            [
                "plot-summary",
                str(out_dir),
                "-f",
                str(dir / "fingerprints.npy"),
                "--no-show",
                "--no-verbose",
            ],
        )
    assert result.exit_code == 0


def test_fps_from_smiles() -> None:
    runner = CliRunner()
    smiles_path = Path(__file__).parent / "chembl-sample-3k.smi"
    with tempfile.TemporaryDirectory() as d:
        dir = Path(d).resolve()
        out_dir = dir / "output"
        result = runner.invoke(
            app, ["fps-from-smiles", str(smiles_path), "-o", str(out_dir)]
        )
        assert result.exit_code == 0
        file = list(out_dir.glob("*.npy"))[0]
        fps_raw = np.load(file).reshape(-1)
    nonzero = fps_raw.nonzero()[0].reshape(-1)
    actual = fps_raw[nonzero][:19].tolist()
    expect = [4, 128, 2, 16, 8, 16, 4, 16, 128, 16, 1, 128, 1, 64, 1, 1, 128, 32, 32]
    assert actual == expect


# NOTE: This is an acceptance test only
def test_fps_from_smiles_invalid() -> None:
    runner = CliRunner()
    smiles_path = Path(__file__).parent / "chembl-sample-bad.smi"
    with tempfile.TemporaryDirectory() as d:
        dir = Path(d).resolve()
        out_dir = dir / "output"
        result = runner.invoke(
            app,
            [
                "fps-from-smiles",
                str(smiles_path),
                "-o",
                str(out_dir),
                "--sanitize",
                "minimal",
                "--skip-invalid",
            ],
        )
        assert result.exit_code == 0
        file = list(out_dir.glob("*fps*.npy"))[0]
        fps_raw = np.load(file).reshape(-1)
    nonzero = fps_raw.nonzero()[0].reshape(-1)
    actual = fps_raw[nonzero][:19].tolist()
    expect = [2, 4, 32, 1, 2, 128, 4, 128, 32, 32, 80, 128, 64, 128, 1, 16, 64, 4, 16]
    assert actual == expect


def test_info() -> None:
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as d:
        dir = Path(d).resolve()
        fps = make_fake_fingerprints(
            3000, n_features=2048, seed=12620509540149709235, pack=True
        )
        np.save(dir / "fingerprints.npy", fps)
        result = runner.invoke(app, ["fps-info", str(dir)])
    if isinstance(result.output, bytes):
        out = result.output.decode("utf-8")
    else:
        out = result.output
    assert result.exit_code == 0
    assert "Valid fingerprint file" in out
    assert "256" in out
    assert "uint8" in out


def test_help() -> None:
    runner = CliRunner()
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "CLI tool for serial or parallel fast clustering" in result.output


def test_multiround() -> None:
    runner = CliRunner()
    with tempfile.TemporaryDirectory() as d:
        dir = Path(d).resolve()
        for seed in range(1, 21, 2):
            fps = make_fake_fingerprints(100, seed=seed)
            np.save(dir / f"fps.{str(seed).zfill(4)}.npy", fps)
        out_dir = dir / "output"
        out_dir.mkdir()
        result = runner.invoke(
            app,
            [
                "multiround",
                str(dir),
                "-o",
                str(out_dir),
                "-t",
                "0.65",
                "--bin-size",
                "2",
                "--ps",
                "1",
                "--no-verbose",
                "--set-mid-merge",
                "tolerance-legacy",
                "--no-monitor-mem",
            ],
        )
        with open(out_dir / "clusters.pkl", mode="rb") as f:
            obj = pickle.load(f)
    assert result.exit_code == 0
    assert EXPECT_MULTIROUND_CLUSTERS == obj[: len(EXPECT_MULTIROUND_CLUSTERS)]


def test_run() -> None:
    runner = CliRunner()
    ids_expect = [
        [2195, 2196, 2378, 2440, 2443, 2454, 2463, 2464, 2465, 2467, 2527, 2544],
        [199, 228, 255, 270, 273, 438, 457, 458, 461, 470, 477, 496],
        [700, 728, 773, 798, 825, 891, 919, 962, 963, 968, 998],
        [1448, 1567, 1590, 1606, 1612, 1637, 1640, 1648, 1686, 1694],
        [1059, 1065, 1072, 1077, 1154, 1194, 1301],
        [1779, 1802, 1807, 1828, 1856, 1864],
        [2826, 2896, 2970, 2973, 2975],
        [1986, 2107, 2139, 2141],
        [1933, 1949],
        [2233, 2294],
        [1551, 1552],
        [1219, 1226],
        [614, 637],
    ]
    with tempfile.TemporaryDirectory() as d:
        Path
        dir = Path(d).resolve()
        fps = make_fake_fingerprints(
            3000, n_features=2048, seed=12620509540149709235, pack=True
        )
        np.save(dir / "fingerprints.npy", fps)
        out_dir = dir / "output"
        result = runner.invoke(
            app,
            [
                "run",
                str(dir),
                "-o",
                str(out_dir),
                "-b",
                "50",
                "-t",
                "0.65",
                "--no-monitor-mem",
            ],
        )
        with open(out_dir / "clusters.pkl", mode="rb") as f:
            obj = pickle.load(f)
        assert result.exit_code == 0
        assert obj[:13] == ids_expect


EXPECT_MULTIROUND_CLUSTERS = [
    [
        368,
        414,
        422,
        423,
        520,
        549,
        581,
        609,
        625,
        683,
        622,
        709,
        761,
        770,
        789,
        813,
        831,
        989,
    ],
    [23, 285, 209, 213, 276, 294, 316, 319, 358],
    [1],
    [2],
    [3],
    [4],
    [5],
    [6],
    [7],
    [8],
    [9],
    [10],
    [11],
    [12],
    [13],
    [14],
    [15],
    [16],
    [17],
    [18],
    [19],
    [20],
    [21],
    [22],
    [24],
    [25],
    [26],
    [27],
    [28],
    [29],
    [30],
    [31],
    [32],
    [33],
    [34],
    [35],
    [36],
    [37],
    [38],
    [39],
    [40],
    [41],
    [42],
    [43],
    [44],
    [45],
    [46],
    [47],
    [48],
    [49],
    [50],
    [51],
    [52],
    [53],
    [54],
    [55],
    [56],
    [57],
    [58],
    [59],
    [60],
    [61],
    [62],
    [63],
    [64],
    [65],
    [66],
    [67],
    [68],
    [69],
    [70],
    [71],
    [72],
    [73],
    [74],
    [75],
    [76],
    [77],
    [78],
    [79],
    [80],
    [81],
    [82],
    [83],
    [84],
    [85],
    [86],
    [87],
    [88],
    [89],
    [90],
    [91],
    [92],
    [94],
    [95],
    [96],
    [97],
    [98],
    [99],
    [0],
    [101],
    [102],
    [103],
    [104],
    [105],
    [106],
    [107],
    [108],
    [109],
    [110],
    [111],
    [112],
    [113],
    [114],
    [115],
    [116],
    [117],
    [118],
    [119],
    [121],
    [122],
    [123],
    [124],
    [125],
    [126],
    [127],
    [128],
    [129],
    [130],
    [131],
]



# type: ignore
import numpy as np


def jt_isim(c_total, n_objects):
    sum_kq = np.sum(c_total)
    sum_kqsq = np.dot(c_total, c_total)
    a = (sum_kqsq - sum_kq) / 2

    return a / (a + n_objects * sum_kq - sum_kqsq)


def merge_tolerance(
    threshold,
    new_ls,
    new_centroid,
    new_n,
    old_ls,
    nom_ls,
    old_n,
    nom_n,
    tolerance=0.05,
):
    jt_radius = jt_isim(new_ls, new_n)
    if jt_radius < threshold:
        return False
    else:
        if old_n == 1 and nom_n == 1:
            return True
        elif nom_n == 1:
            return (
                jt_isim(old_ls + nom_ls, old_n + 1) * (old_n + 1)
                - jt_isim(old_ls, old_n) * (old_n - 1)
            ) / 2 >= jt_isim(old_ls, old_n) - tolerance and (jt_radius >= threshold)
        else:
            return True


def merge_radius(
    threshold,
    new_ls,
    new_centroid,
    new_n,
    old_ls,
    nom_ls,
    old_n,
    nom_n,
):
    jt_sim = jt_isim(new_ls + new_centroid, new_n + 1) * (new_n + 1) - jt_isim(
        new_ls, new_n
    ) * (new_n - 1)
    return jt_sim >= threshold * 2


def merge_diameter(
    threshold,
    new_ls,
    new_centroid,
    new_n,
    old_ls,
    nom_ls,
    old_n,
    nom_n,
):
    jt_radius = jt_isim(new_ls, new_n)
    return jt_radius >= threshold



r"""Merging criteria for BitBIRCH clustering

The functionality in this module is advanced and not needed for normal usage of the
library. Make sure you understand the features here before applying them.
"""

import re
import typing as tp
import numpy as np
from numpy.typing import NDArray

# NOTE: jt_isim_from_sum is equivalent to jt_isim_diameter_compl_from_sum
from bblean.similarity import jt_isim_from_sum, jt_isim_radius_compl_from_sum

_BUILTIN_MERGES = [
    "radius",
    "diameter",
    "flexible-tolerance-diameter",
    "tolerance-diameter",
    "tolerance-radius",
    "tolerance",
    "never",
]


class DiscardSubcluster(Exception):
    r"""If raised in hooks, immediatly exit the merge discarding the incident subcluster

    Discarded subclusters will not be stored in the final tree, and will only show
    up if calling `bblean.BitBirch.get_assigments` (or the `labels_` attribute if
    using `bblean.sklearn`) with a cluster label of 0.
    """


class RejectMerge(Exception):
    r"""If raised in hooks, immediatly exit the merge and reject it"""


class MergeAcceptFunction:
    r"""Base class for user defined merges

    If you want to implement a custom BitBirch merge you can subclass this and
    pass an instance of this function to a `bblean.BitBirch` class upon creation
    as ``BitBirch(..., merge_criterion=instance)``.

    .. warning::

        This is an advanced feature, make sure you fully understand what you are doing!
    """

    # For the merge functions, although outputs of jt_isim_from_sum f64, directly using
    # f64 is *not* faster than starting with uint64

    def __call__(
        self,
        thresh: float,
        new_ls: NDArray[np.integer],
        new_n: int,
        old_ls: NDArray[np.integer],
        nom_ls: NDArray[np.integer],
        old_n: int,
        nom_n: int,
        old_idxs: tp.Sequence[int],
        nom_idxs: tp.Sequence[int],
    ) -> bool:
        try:
            thresh = self.on_check_merge_start(
                thresh, new_ls, new_n, old_ls, nom_ls, old_n, nom_n, old_idxs, nom_idxs
            )
            accepted = self.check_merge(
                thresh, new_ls, new_n, old_ls, nom_ls, old_n, nom_n, old_idxs, nom_idxs
            )
            self.on_check_merge_end(accepted, old_idxs, nom_idxs)
        except RejectMerge:
            return False
        return accepted

    def on_check_merge_start(
        self,
        threshold: float,
        new_sum: NDArray[np.integer],
        new_n: int,
        old_sum: NDArray[np.integer],
        nominee_sum: NDArray[np.integer],
        old_n: int,
        nominee_n: int,
        old_idxs: tp.Sequence[int],
        nominee_idxs: tp.Sequence[int],
    ) -> float:
        r"""Hook called before a merge is checked (meant to be overriden)

        See `MergeAcceptFunction.check_merge` for an explanation of the different args

        .. warning::

            Numpy arrays passed to this function may use uint types, watch out for
            pitfalls of unsigned integer arithmetic.

        This function must return the threshold, unchanged. If the threshold is modified
        by this function, the new threshold will be used for this specific merge check.
        """
        return threshold

    def check_merge(
        self,
        threshold: float,
        new_sum: NDArray[np.integer],
        new_n: int,
        old_sum: NDArray[np.integer],
        nominee_sum: NDArray[np.integer],
        old_n: int,
        nominee_n: int,
        old_idxs: tp.Sequence[int],
        nominee_idxs: tp.Sequence[int],
    ) -> bool:
        r"""Check if a merge should be accepted.

        All user-defined merges should subclass this function.

        threshold: Threshold for the merge
        new_sum: old_sum + nominee_sum
        new_n: old_n + new_n
        old_sum: col-wise sum of all fingerprints in this cluster
        nominee_sum: col-wise sum of all fingerprints in the nominee cluster
        old_n: size of this cluster
        nominee_n: size of the nominee cluster
        old_idxs: Indices of the mols in cluster before the nominee cluster is merged
        nominee_idxs: Nominee indices to merge.

        If merging a single molecule (the most common case, when calling
        `bblean.BitBirch.fit`), ``nominee_idxs`` will be a list with a *single index*,
        ``new_n = 1``, and ``nominee_sum`` will be the molecule fingerprint.

        This function must return a boolean that determines whether the merge was
        accepted

        .. warning::

            Numpy arrays passed to this function may use uint types, watch out for
            pitfalls of unsigned integer arithmetic.
        """
        raise NotImplementedError

    def on_check_merge_end(
        self,
        accepted: bool,
        old_idxs: tp.Sequence[int],
        nominee_idxs: tp.Sequence[int],
    ) -> None:
        r"""Hook called after a merge is checked (meant to be overriden)

        accept: Whether the merge was accepted
        old_idxs: Indices of the mols in cluster before the nominee cluster is merged
        nominee_idxs: Nominee indices to merge.

        If merging a single molecule (the most common case, when calling
        `bblean.BitBirch.fit`, ``nominee_idxs`` will be a list with a *single index*)

        This function must not return a value
        """

    @property
    def name(self) -> str:
        return "-".join(
            s.lower()
            for s in re.split(r"(?=[A-Z])", self.__class__.__name__)[1:]
            if s != "Merge"
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class RadiusMerge(MergeAcceptFunction):

    def check_merge(
        self,
        threshold: float,
        new_sum: NDArray[np.integer],
        new_n: int,
        old_sum: NDArray[np.integer],
        nominee_sum: NDArray[np.integer],
        old_n: int,
        nominee_n: int,
        old_idxs: tp.Sequence[int] | None = None,
        nominee_idxs: tp.Sequence[int] | None = None,
    ) -> bool:
        r""":meta private:"""
        return jt_isim_radius_compl_from_sum(new_sum, new_n) >= threshold


class DiameterMerge(MergeAcceptFunction):

    def check_merge(
        self,
        threshold: float,
        new_sum: NDArray[np.integer],
        new_n: int,
        old_sum: NDArray[np.integer],
        nominee_sum: NDArray[np.integer],
        old_n: int,
        nominee_n: int,
        old_idxs: tp.Sequence[int] | None = None,
        nominee_idxs: tp.Sequence[int] | None = None,
    ) -> bool:
        r""":meta private:"""
        return jt_isim_from_sum(new_sum, new_n) >= threshold


class FlexibleToleranceDiameterMerge(MergeAcceptFunction):
    # NOTE: Equivalent to tolerance-diameter but uses min(old_dc, threshold) as the
    # criteria

    def __init__(
        self,
        tolerance: float = 0.05,
        n_max: int = 1000,
        decay: float = 1e-3,
        adaptive: bool = True,
    ) -> None:
        self.tolerance = tolerance
        self.decay = decay
        self.offset = np.exp(-decay * n_max)
        if not adaptive:
            self.decay = 0.0
            self.offset = 0.0

    def check_merge(
        self,
        threshold: float,
        new_sum: NDArray[np.integer],
        new_n: int,
        old_sum: NDArray[np.integer],
        nominee_sum: NDArray[np.integer],
        old_n: int,
        nominee_n: int,
        old_idxs: tp.Sequence[int] | None = None,
        nominee_idxs: tp.Sequence[int] | None = None,
    ) -> bool:
        r""":meta private:"""
        new_dc = jt_isim_from_sum(new_sum, new_n)
        if new_dc < threshold:
            return False
        if old_n == 1:
            return True
        old_dc = jt_isim_from_sum(old_sum, old_n)
        tol = max(self.tolerance * (np.exp(-self.decay * old_n) - self.offset), 0.0)
        return new_dc >= min(old_dc, threshold) - tol

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.tolerance})"


class ToleranceDiameterMerge(MergeAcceptFunction):
    # NOTE: The reliability of the estimate of the cluster should be a function of the
    # size of the old cluster, so in this metric, tolerance is larger for small clusters
    # tolerance = max{ alpha * (exp(-decay * N_old) - offset), 0}

    def __init__(
        self,
        tolerance: float = 0.05,
        n_max: int = 1000,
        decay: float = 1e-3,
        adaptive: bool = True,
    ) -> None:
        self.tolerance = tolerance
        self.decay = decay
        self.offset = np.exp(-decay * n_max)
        if not adaptive:
            self.decay = 0.0
            self.offset = 0.0

    def check_merge(
        self,
        threshold: float,
        new_sum: NDArray[np.integer],
        new_n: int,
        old_sum: NDArray[np.integer],
        nominee_sum: NDArray[np.integer],
        old_n: int,
        nominee_n: int,
        old_idxs: tp.Sequence[int] | None = None,
        nominee_idxs: tp.Sequence[int] | None = None,
    ) -> bool:
        r""":meta private:"""
        new_dc = jt_isim_from_sum(new_sum, new_n)
        if new_dc < threshold:
            return False
        # If the old n is 1 then merge directly (infinite tolerance), since the
        # old_d is undefined for a single fp
        if old_n == 1:
            return True
        # Only merge if the new_dc is greater or equal to the old, up to some tolerance,
        # which decays with N
        old_dc = jt_isim_from_sum(old_sum, old_n)
        tol = max(self.tolerance * (np.exp(-self.decay * old_n) - self.offset), 0.0)
        return new_dc >= old_dc - tol

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.tolerance})"


class ToleranceRadiusMerge(ToleranceDiameterMerge):

    def check_merge(
        self,
        threshold: float,
        new_sum: NDArray[np.integer],
        new_n: int,
        old_sum: NDArray[np.integer],
        nominee_sum: NDArray[np.integer],
        old_n: int,
        nominee_n: int,
        old_idxs: tp.Sequence[int] | None = None,
        nominee_idxs: tp.Sequence[int] | None = None,
    ) -> bool:
        r""":meta private:"""
        new_rc = jt_isim_radius_compl_from_sum(new_sum, new_n)
        if new_rc < threshold:
            return False
        if old_n == 1:
            return True
        old_rc = jt_isim_radius_compl_from_sum(old_sum, old_n)
        tol = max(self.tolerance * (np.exp(-self.decay * old_n) - self.offset), 0.0)
        return new_rc >= old_rc - tol

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.tolerance})"


class NeverMerge(ToleranceDiameterMerge):
    r""":meta private:"""

    def check_merge(
        self,
        threshold: float,
        new_sum: NDArray[np.integer],
        new_n: int,
        old_sum: NDArray[np.integer],
        nominee_sum: NDArray[np.integer],
        old_n: int,
        nominee_n: int,
        old_idxs: tp.Sequence[int] | None = None,
        nominee_idxs: tp.Sequence[int] | None = None,
    ) -> bool:
        return False

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class ToleranceLegacyMerge(MergeAcceptFunction):
    r""":meta private:"""

    def __init__(self, tolerance: float = 0.05) -> None:
        self.tolerance = tolerance

    def check_merge(
        self,
        threshold: float,
        new_sum: NDArray[np.integer],
        new_n: int,
        old_sum: NDArray[np.integer],
        nominee_sum: NDArray[np.integer],
        old_n: int,
        nominee_n: int,
        old_idxs: tp.Sequence[int] | None = None,
        nominee_idxs: tp.Sequence[int] | None = None,
    ) -> bool:
        # First two branches are equivalent to 'diameter'
        new_dc = jt_isim_from_sum(new_sum, new_n)
        if new_dc < threshold:
            return False
        if old_n == 1 or nominee_n != 1:
            return True
        # 'new_dc >= threshold' and 'new_n == old_n + 1' are guaranteed here
        old_dc = jt_isim_from_sum(old_sum, old_n)
        return (new_dc * new_n - old_dc * (old_n - 1)) / 2 >= old_dc - self.tolerance

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.tolerance})"


# Make these ones leaner with less calls so they don't take up any extra time at all in
# the default case
class _FastMerge(MergeAcceptFunction):
    def __call__(
        self,
        thresh: float,
        new_ls: NDArray[np.integer],
        new_n: int,
        old_ls: NDArray[np.integer],
        nom_ls: NDArray[np.integer],
        old_n: int,
        nom_n: int,
        old_idxs: tp.Sequence[int],
        nom_idxs: tp.Sequence[int],
    ) -> bool:
        return self.check_merge(
            thresh, new_ls, new_n, old_ls, nom_ls, old_n, nom_n, old_idxs, nom_idxs
        )

    @property
    def name(self) -> str:
        return "-".join(
            s.lower()
            for s in re.split(r"(?=[A-Z])", self.__class__.__name__)[1:]
            if s not in ["Merge", "Fast"]
        )


class _FastDiameterMerge(_FastMerge, DiameterMerge):
    pass


class _FastRadiusMerge(_FastMerge, RadiusMerge):
    pass


class _FastToleranceDiameterMerge(_FastMerge, ToleranceDiameterMerge):
    pass


class _FastFlexibleToleranceDiameterMerge(_FastMerge, FlexibleToleranceDiameterMerge):
    pass


class _FastToleranceRadiusMerge(_FastMerge, ToleranceRadiusMerge):
    pass


class _FastToleranceLegacyMerge(_FastMerge, ToleranceLegacyMerge):
    pass


class _FastNeverMerge(_FastMerge, NeverMerge):
    pass


def _get_merge_accept_fn(
    merge_criterion: str, tolerance: float = 0.05
) -> MergeAcceptFunction:
    if merge_criterion == "radius":
        return _FastRadiusMerge()
    elif merge_criterion == "diameter":
        return _FastDiameterMerge()
    elif merge_criterion == "tolerance-legacy":
        return _FastToleranceLegacyMerge(tolerance)
    elif merge_criterion == "tolerance-diameter":
        return _FastToleranceDiameterMerge(tolerance)
    elif merge_criterion == "flexible-tolerance-diameter":
        return _FastFlexibleToleranceDiameterMerge(tolerance)
    elif merge_criterion == "tolerance-radius":
        return _FastToleranceRadiusMerge(tolerance)
    elif merge_criterion == "never":
        return _FastNeverMerge(tolerance)
    raise ValueError(
        f"Unknown merge criterion {merge_criterion} "
        "Valid criteria are: radius|diameter|tolerance-diameter|tolerance-radius"
    )



# type: ignore
import numpy as np


def intra_sim(clust1, clust2):
    """Similarity between clusters from the iSIM of their union"""
    n1 = len(clust1)
    n2 = len(clust2)
    combined = np.sum(clust1, axis=0) + np.sum(clust2, axis=0)
    return jt_isim(combined, n1 + n2)


def calculate_comp_sim(data):
    """Returns vector of complementary similarities"""
    n_objects = len(data) - 1
    c_total = np.sum(data, axis=0)
    comp_matrix = c_total - data
    a = comp_matrix * (comp_matrix - 1) / 2
    comp_sims = np.sum(a, axis=1) / np.sum(
        (a + comp_matrix * (n_objects - comp_matrix)), axis=1
    )
    return comp_sims


def calculate_medoid(data):
    """Returns index of medoid"""
    return np.argmin(calculate_comp_sim(data))


def jt_isim(c_total, n_objects):
    """iSIM Tanimoto calculation"""
    sum_kq = np.sum(c_total)
    sum_kqsq = np.dot(c_total, c_total)
    a = (sum_kqsq - sum_kq) / 2
    return a / (a + n_objects * sum_kq - sum_kqsq)


def jt_pair(mol1, mol2):
    """Tanimoto between two molecules"""
    a = np.dot(mol1, mol2)
    return a / (np.sum(mol1) + np.sum(mol2) - a)


def jt_one_to_many(mol, mol_set):
    """Tanimoto similarities between a molecule and a set of molecules

    mol: np.array of a single molecule
    mol_set: np.array containing the fingerprints of a set of molecules
    """
    pop_counts = np.sum(mol_set, axis=1)
    a = np.dot(mol_set, mol)
    sims = a / (np.sum(mol) + pop_counts - a)
    return sims


def calculate_centroid(linear_sum, n_samples):
    """Calculates centroid"""
    return np.floor(linear_sum / n_samples + 0.5)


def legacy_chi(clusters, reps=False, rep_type="centroid"):
    """Calinski-Harabasz index

    clusters : list of np.arrays containing the clusters

    reps : {bool, list} indicates if cluster representatives are given

    rep_type : type of representative, medoid or centroid

    curated : bool indicates if singletons have been removed

    min_size : int size below which clusters will be ignored

    Note
    ----
    Higher values are better

    """
    # n : total number of points
    n = 0

    # k : number of clusters
    k = len(clusters)

    # wcss : within-cluster sum of squares
    wcss = 0

    # bcss : between-cluster sum of squares
    bcss = 0
    total_data = []
    for clust in clusters:
        for mol in clust:
            total_data.append(mol)
    total_data = np.array(total_data)
    if rep_type == "centroid":
        linear_sum = np.sum(total_data, axis=0)
        n_samples = len(total_data)
        c = calculate_centroid(linear_sum, n_samples)
    elif rep_type == "medoid":
        medoid = calculate_medoid(total_data)
        c = total_data[medoid]
    if not reps:
        for clust in clusters:
            n_samples = len(clust)
            n += n_samples
            if rep_type == "centroid":
                linear_sum = np.sum(clust, axis=0)
                rep = calculate_centroid(linear_sum, n_samples)
            elif rep_type == "medoid":
                medoid = calculate_medoid(clust)
                rep = clust[medoid]
            distances = 1 - jt_one_to_many(rep, clust)
            wcss += np.dot(distances, distances)
            # NOTE: Fix fatal bug here, original impl lacks 1 - ...
            bcss += n_samples * (1 - jt_pair(c, rep)) ** 2
    else:
        for i, clust in enumerate(clusters):
            n_samples = len(clust)
            n += n_samples
            # NOTE: Fix fatal bug here, original impl lacks 1 - ...
            bcss += n_samples * (1 - jt_pair(c, reps[i])) ** 2
            distances = 1 - jt_one_to_many(reps[i], clust)
            wcss += np.dot(distances, distances)

    try:
        value = bcss * (n - k) / (wcss * (k - 1))
    except Exception:
        value = 0
    return value


def legacy_dbi(clusters, reps=False, rep_type="centroid"):
    """Davies-Bouldin index

    clusters : list of np.arrays containing the clusters

    reps : {bool, list} indicates if cluster representatives are given

    rep_type : type of representative, medoid or centroid

    curated : bool indicates if singletons have been removed

    min_size : int size below which clusters will be ignored

    Note
    ----
    Lower values are better

    """
    n = 0

    S = []

    if not reps:
        reps = []
        for clust in clusters:
            n_samples = len(clust)
            n += n_samples
            if rep_type == "centroid":
                linear_sum = np.sum(clust, axis=0)
                rep = calculate_centroid(linear_sum, n_samples)
            elif rep_type == "medoid":
                medoid = calculate_medoid(clust)
                rep = clust[medoid]
            reps.append(rep)
            S.append(np.sum(1 - jt_one_to_many(rep, clust)) / n_samples)
    else:
        for i, clust in enumerate(clusters):
            n_samples = len(clust)
            n += n_samples
            S.append(np.sum(1 - jt_one_to_many(reps[i], clust)) / n_samples)

    db = 0

    for i, clust in enumerate(clusters):
        d = []
        for j, other_clust in enumerate(clusters):
            if i == j:
                d.append(-1)
            else:
                Mij = 1 - jt_pair(reps[i], reps[j])
                Rij = (S[i] + S[j]) / Mij
                d.append(Rij)
        db += max(d)

    try:
        value = db / n
    except Exception:
        value = 0

    return value


def legacy_dunn(
    clusters,
    cluster_sim="intra",
    reps=False,
    rep_type="centroid",
):
    """Dunn index

    cluster_sim : {'intra', 'inter'} type of similarity between clusters
        intra : intra_sim
        inter : inter_sim

    clusters : list of np.arrays containing the clusters

    reps : {bool, list} indicates if cluster representatives are given

    rep_type : type of representative, medoid or centroid

    curated : bool indicates if singletons have been removed

    min_size : int size below which clusters will be ignored

    Note
    ----
    Higher values are better

    """
    D = []

    for clust in clusters:
        n_samples = len(clust)
        linear_sum = np.sum(clust, axis=0)
        D.append(jt_isim(linear_sum, n_samples))

    if len(D) == 0:
        Dm = 0
    else:
        Dm = max(D)

    # initial min_d value could be any number > 1
    min_d = 3.08

    for i, clust1 in enumerate(clusters):
        for j, clust2 in enumerate(clusters):
            if i == j:
                pass
            else:
                dij = 1 - intra_sim(clust1, clust2)
                if dij < min_d:
                    min_d = dij

    try:
        value = min_d / Dm
    except Exception:
        value = 0

    return value



r"""BitBirch 'Lean' classes that fully respects the sklearn API contract.

Use these classes as a drop-in replacement of `sklearn.cluster.Birch` if you are used to
the `sklearn` way of doing things, with the caveat that global clustering is not
currently supported.
"""

import typing as tp
from numpy.typing import NDArray
import numpy as np
import typing_extensions as tpx

from sklearn.utils.validation import check_is_fitted, validate_data
from sklearn.metrics import pairwise_distances_argmin, pairwise_distances
from sklearn.base import (
    BaseEstimator,
    ClassNamePrefixFeaturesOutMixin,
    ClusterMixin,
    TransformerMixin,
    _fit_context,
)

from bblean.fingerprints import unpack_fingerprints
from bblean.bitbirch import BitBirch as _BitBirch
from bblean.merges import MergeAcceptFunction

__all__ = ["BitBirch", "UnpackedBitBirch"]

# Required functions for sklearn API:
# - fit() *must be defined*
# - transform()  *must be defined*
# - fit_predict()  (ClusterMixin) default implementation is to fit and then return lbls
# - predict()  # overloaded to use jt instead of euclidean
# - fit_transform()  (TransformerMixin, delegates to *fit* and *transform*)
# - set_output()  (TransformerMixin via _SetOutputMixin)
# set_output(transform="pandas") or transform="default" (numpy array) (or "polars",
# if polars is installed)

#  The following requires _n_features_out after fitting
# - get_feature_names_out() ["bitbirch0", "bitbirch1", ...] (ClassNamePrefix...)

# - get_metadata_routing() () (!?) New feature, unclear what this is and unnecessary
# - partial_fit() ()  Same as fit() for BitBirch

# These require that the parameters are specified in __init__, and are assigned
# to names (or attributes) with the convention self.<param>.
# - get_params() (BaseEstimator)
# - set_params() (BaseEstimator)


class BitBirch(
    ClassNamePrefixFeaturesOutMixin,
    ClusterMixin,
    TransformerMixin,
    BaseEstimator,
    _BitBirch,
):
    r"""Implements the BitBIRCH clustering algorithm, 'Lean' version.

    Inputs to this estimator are *packed* fingerprints by default. If you need get a
    class that always accepts an unpacked input use `bblean.sklearn.UnpackedBitBirch`

    See `bblean.bitbirch.BitBirch` for more details"""

    _parameter_constraints: dict[str, list[tp.Any]] = {}

    def __init__(
        self,
        *,
        threshold: float = 0.65,
        branching_factor: int = 50,
        merge_criterion: str | MergeAcceptFunction | None = None,
        tolerance: float | None = None,
        compute_labels: bool = True,
    ):
        super().__init__(
            threshold=threshold,
            branching_factor=branching_factor,
            merge_criterion=merge_criterion,
            tolerance=tolerance,
        )
        self.compute_labels = compute_labels

    @_fit_context(prefer_skip_nested_validation=True)
    def fit(  # type: ignore
        self, X, y=None, input_is_packed: bool = True, n_features: int | None = None
    ) -> tpx.Self:
        super().fit(X, input_is_packed=input_is_packed, n_features=n_features)
        centroids = np.stack(
            [bf.unpacked_centroid for bf in self._get_leaf_bfs(sort=True)]
        )
        self.subcluster_centers_ = centroids
        self.subcluster_labels_ = np.arange(1, len(centroids) + 1)
        self._n_features_out = centroids.shape[0]
        if self.compute_labels:
            self.labels_ = self.get_assignments()
        return self

    @_fit_context(prefer_skip_nested_validation=True)
    def partial_fit(  # type: ignore
        self,
        X=None,
        y=None,
        input_is_packed: bool = True,
        n_features: int | None = None,
    ) -> tpx.Self:
        if X is None:
            raise ValueError()
        self.fit(X, input_is_packed=input_is_packed, n_features=n_features)
        if self.compute_labels:
            self.labels_ = self.get_assignments()
        return self

    # Overloaded since self.labels_ may not be set
    def fit_predict(  # type: ignore
        self, X, y=None, input_is_packed: bool = True, n_features: int | None = None
    ) -> NDArray[np.integer]:
        self.fit(X, input_is_packed=input_is_packed, n_features=n_features)
        if not self.compute_labels:
            self.labels_ = self.get_assignments()
        return self.labels_

    def predict(  # type: ignore
        self, X, input_is_packed: bool = True, n_features: int | None = None
    ) -> NDArray[np.integer]:
        """Predict data using the ``centroids`` of subclusters."""
        check_is_fitted(self)
        X = validate_data(self, X, accept_sparse="csr", reset=False)
        X = (
            (unpack_fingerprints(X, n_features=n_features) if input_is_packed else X)
            .astype(np.uint8, copy=False)
            .view(np.bool)
        )
        # TODO: Due to a sklearn bug this performs unnecessary casts
        centers = self.subcluster_centers_.astype(np.uint8, copy=False).view(np.bool)
        argmin = pairwise_distances_argmin(X, centers, metric="jaccard")
        return self.subcluster_labels_[argmin]

    def transform(  # type: ignore
        self,
        X,
        input_is_packed: bool = True,
        n_features: int | None = None,
    ):
        check_is_fitted(self)
        X = validate_data(self, X, accept_sparse="csr", reset=False)
        X = (
            (unpack_fingerprints(X, n_features=n_features) if input_is_packed else X)
            .astype(np.uint8, copy=False)
            .view(np.bool)
        )
        centers = self.subcluster_centers_.astype(np.uint8, copy=False).view(np.bool)
        return pairwise_distances(X, centers, metric="jaccard")

    def __sklearn_tags__(self):  # type: ignore
        tags = super().__sklearn_tags__()
        tags.input_tags.sparse = True
        return tags


class UnpackedBitBirch(BitBirch):
    r"""Implements the BitBIRCH clustering algorithm, 'Lean' version.

    Inputs to this estimator are *unpacked* fingerprints always

    See `bblean.bitbirch.BitBirch` for more details"""

    def fit(  # type: ignore
        self, X, y=None, input_is_packed: bool = False, n_features: int | None = None
    ) -> tpx.Self:
        return super().fit(X, y, input_is_packed=input_is_packed, n_features=n_features)

    def partial_fit(  # type: ignore
        self, X, y=None, input_is_packed: bool = False, n_features: int | None = None
    ):
        return super().partial_fit(
            X, y, input_is_packed=input_is_packed, n_features=n_features
        )

    def fit_predict(  # type: ignore
        self, X, y=None, input_is_packed: bool = False, n_features: int | None = None
    ):
        return super().fit_predict(
            X, y, input_is_packed=input_is_packed, n_features=n_features
        )

    def predict(  # type: ignore
        self,
        X,
        input_is_packed: bool = False,
        n_features: int | None = None,
    ):
        return super().predict(
            X, input_is_packed=input_is_packed, n_features=n_features
        )

    def transform(  # type: ignore
        self, X, input_is_packed: bool = False, n_features: int | None = None
    ):
        return super().transform(
            X, input_is_packed=input_is_packed, n_features=n_features
        )



r"""Monitor and collect memory stats"""

import typing as tp
import mmap
import warnings
from enum import Enum
import ctypes
import dataclasses
from pathlib import Path
import sys
import time
import os
import multiprocessing as mp

import typing_extensions as tpx
import psutil
import numpy as np
from numpy.typing import NDArray
from rich.console import Console

_BYTES_TO_GIB = 1 / 1024**3


class Madv(Enum):
    WILLNEED = 3
    SEQUENTIAL = 2
    # PAGEOUT and DONTNEED reduce memory usage around 40%
    # TODO: Check exactly what DONTNEED does. I believe PAGEOUT *swaps out*
    # so DONTNEED may be preferred since it may have less perf. penalty
    DONTNEED = 4
    PAGEOUT = 21
    FREE = 8  # *ONLY* works on anonymous pages (not file-backed like numpy arrays)
    # Cold does *not* immediatly release memory, it is just a soft hint that
    # those pages won't be needed soon
    COLD = 20


# Get handle to the system's libc
def _get_libc() -> tp.Any:
    if sys.platform == "linux":
        return ctypes.CDLL("libc.so.6", use_errno=True)
    elif sys.platform == "darwin":
        return ctypes.CDLL("libc.dylib", use_errno=True)
    # For now, do nothing in Windows
    return


# This reduces memory usage around 40%, since the kernel can release
# pages once the array has been iterated over. The issue is, after this has been done,
# the array is out of the RAM, so refinement is not possible.
def _madvise_dontneed(page_start: int, size: int) -> None:
    _madvise(page_start, size, Madv.DONTNEED)


# let the kernel know that access to this range of addrs will be sequential
# (pages can be read-ahead and discarded fast after read if needed)
def _madvise_sequential(page_start: int, size: int) -> None:
    _madvise(page_start, size, Madv.SEQUENTIAL)


def _madvise(page_start: int, size: int, opt: Madv) -> None:
    libc = _get_libc()
    if libc is None:
        return
    if libc.madvise(ctypes.c_void_p(page_start), size, opt.value) != 0:
        errno = ctypes.get_errno()
        warnings.warn(f"{opt} failed with error code {errno}")


_Input = tp.Union[NDArray[np.integer], list[NDArray[np.integer]]]


@dataclasses.dataclass
class _ArrayMemPagesManager:
    can_release: bool
    _pagesizex: int
    _iters_per_pagex: int
    _curr_page_start_addr: int

    @classmethod
    def from_bb_input(cls, X: _Input, can_release: bool | None = None) -> tpx.Self:
        pagesizex = mmap.PAGESIZE * 512
        if (
            isinstance(X, np.memmap)
            and X.ndim == 2
            and (pagesizex % X.shape[1] == 0)
            and X.offset < X.shape[1]
        ):
            # In most cases pagesizex % n_features == 0 and offset < n_features
            # Every n_iters, release the prev page and add pagesizex to start_addr
            iters_per_pagex = int(pagesizex / X.shape[1])  # ~ 8192 iterations
            curr_page_start_addr = X.ctypes.data - X.offset
            _can_release = True
        else:
            iters_per_pagex = 0
            curr_page_start_addr = 0
            _can_release = False
        if can_release is not None:
            _can_release = can_release
        return cls(_can_release, pagesizex, iters_per_pagex, curr_page_start_addr)

    def should_release_curr_page(self, row_idx: int) -> bool:
        return row_idx % self._iters_per_pagex == 0

    def release_curr_page_and_update_addr(self) -> None:
        _madvise_dontneed(self._curr_page_start_addr, self._pagesizex)
        self._curr_page_start_addr += self._pagesizex


def _mmap_file_and_madvise_sequential(
    path: Path, max_fps: int | None = None
) -> NDArray[np.integer]:
    arr = np.load(path, mmap_mode="r")[:max_fps]
    # Numpy actually puts the *whole file* in mmap mode (arr + header)
    # This means the array data starts from a nonzero offset starting from the backing
    # buffer if we want the address to the start of the file we need to displace the
    # addr of the arry by the bsize of the header, which can be accessed by arr.offset
    #
    # This is required since madvise needs a page-aligned address (address must
    # be a multiple of mmap.PAGESIZE (portable) == os.sysconf("SC_PAGE_SIZE")
    # (mac|linux), typically 4096 B).
    #
    # TODO: In some cases, for some reason, this fails with errno 22
    # failure is harmless, but could incurr in a slight perf penalty
    _madvise_sequential(arr.ctypes.data - arr.offset, arr.nbytes)
    return arr


def system_mem_gib() -> tuple[int, int] | tuple[None, None]:
    mem = psutil.virtual_memory()
    return mem.total * _BYTES_TO_GIB, mem.available * _BYTES_TO_GIB


def get_peak_memory_gib(out_dir: Path) -> float | None:
    file = out_dir / "max-rss.json"
    if not file.exists():
        return None
    with open(file, mode="r", encoding="utf-8") as f:
        peak_mem_gib = float(f.read().strip())
    return peak_mem_gib


def monitor_rss_process(
    file: Path | str, interval_s: float, start_time: float, parent_pid: int
) -> None:
    file = Path(file)
    this_pid = os.getpid()
    ps = psutil.Process(parent_pid)

    def total_rss() -> float:
        total_rss = ps.memory_info().rss
        for proc in ps.children(recursive=True):
            if proc.pid == this_pid:
                continue
            try:
                total_rss += proc.memory_info().rss
            except psutil.NoSuchProcess:
                # Prevent race condition since process may have finished before it can
                # be polled
                continue
        return total_rss

    with open(file, mode="w", encoding="utf-8") as f:
        f.write("rss_gib,time_s\n")
        f.flush()
        os.fsync(f.fileno())

    max_rss_gib = 0.0
    while True:
        total_rss_gib = total_rss() * _BYTES_TO_GIB
        with open(file, mode="a", encoding="utf-8") as f:
            f.write(f"{total_rss_gib},{time.perf_counter() - start_time}\n")
            f.flush()
            os.fsync(f.fileno())
        if total_rss_gib > max_rss_gib:
            max_rss_gib = total_rss_gib
            with open(file.parent / "max-rss.json", mode="w", encoding="utf-8") as f:
                f.write(f"{max_rss_gib}\n")
                f.flush()
                os.fsync(f.fileno())
        time.sleep(interval_s)


def launch_monitor_rss_daemon(
    out_file: Path, interval_s: float, console: Console | None = None
) -> None:
    if console is not None:
        console.print("** Monitoring total RAM usage **\n")
    mp.Process(
        target=monitor_rss_process,
        kwargs=dict(
            file=out_file,
            interval_s=interval_s,
            start_time=time.perf_counter(),
            parent_pid=os.getpid(),
        ),
        daemon=True,
    ).start()



from numpy.typing import NDArray
import numpy as np
import pytest

from bblean.fingerprints import make_fake_fingerprints
from bblean.metrics import jt_dbi, jt_isim_chi, jt_isim_dunn
from legacy_metrics import legacy_dunn, legacy_chi, legacy_dbi  # type: ignore


@pytest.fixture
def clusters() -> list[NDArray[np.uint8]]:
    fps = make_fake_fingerprints(
        50, n_features=32, seed=12620509540149709235, pack=True
    )
    clusters = []
    for i in range(5):
        clusters.append(fps[i : i + 10])
    return clusters


@pytest.fixture
def unpacked_clusters() -> list[NDArray[np.uint8]]:
    fps = make_fake_fingerprints(
        50, n_features=32, seed=12620509540149709235, pack=False
    )
    clusters = []
    for i in range(5):
        clusters.append(fps[i : i + 10])
    return clusters


def test_metrics(
    clusters: list[NDArray[np.uint8]], unpacked_clusters: list[NDArray[np.uint8]]
) -> None:
    dbi = jt_dbi(clusters)
    dunn = jt_isim_dunn(clusters)
    chi = jt_isim_chi(clusters)

    expect_dbi = legacy_dbi(unpacked_clusters)
    expect_dunn = legacy_dunn(unpacked_clusters)
    expect_chi = legacy_chi(unpacked_clusters)
    assert dbi == expect_dbi
    assert dunn == expect_dunn
    assert chi == expect_chi
    assert dbi == 0.5183197174528508
    assert dunn == 2.1591579714616436
    assert chi == 4.411591622357663
